from __future__ import absolute_import

from spidermon.contrib.stats.analyzer import StatsAnalyzer
from spidermon.contrib.stats.counters import DictPercentCounter, PercentCounter
from spidermon.exceptions import NotConfigured

from .job import JobMonitorMixin
from .stats import StatsMonitorMixin

DOWNLOADER_RESPONSE_COUNT = 'downloader/response_count'
DOWNLOADER_RESPONSE_STATUS = 'downloader/response_status_count/'
DOWNLOADER_STATUS_CODES_INFORMATIONAL = [r'1\d{2}$']
DOWNLOADER_STATUS_CODES_SUCCESSFUL = [r'2\d{2}$']
DOWNLOADER_STATUS_CODES_REDIRECTIONS = [r'3\d{2}$']
DOWNLOADER_STATUS_CODES_BAD_REQUESTS = [r'4\d{2}$']
DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS = [r'5\d{2}$']
DOWNLOADER_STATUS_CODES_OTHERS = ['[^1-5].*$']
DOWNLOADER_STATUS_CODES_ERRORS = DOWNLOADER_STATUS_CODES_BAD_REQUESTS + \
    DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS


class ResponsesInfo(object):
    def __init__(self, stats):
        self._stats_analyzer = StatsAnalyzer(stats=stats)
        self.count = self._stats_analyzer.search(
            DOWNLOADER_RESPONSE_COUNT + '$'
        ).get(DOWNLOADER_RESPONSE_COUNT, 0)

        # all status codes
        self.all = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=None,
            target=self.all
        )

        # 1xx. informational
        self.informational = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_INFORMATIONAL,
            target=self.informational
        )

        # 2xx. successful
        self.successful = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_SUCCESSFUL,
            target=self.successful
        )

        # 3xx. redirections
        self.redirections = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_REDIRECTIONS,
            target=self.redirections
        )

        # 4xx. bad requests
        self.bad_requests = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_BAD_REQUESTS,
            target=self.bad_requests
        )

        # 5xx. internal server errors
        self.internal_server_errors = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_INTERNAL_SERVER_ERRORS,
            target=self.internal_server_errors
        )

        # >= 6xx. others
        self.others = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_OTHERS, target=self.others)

        # errors (4xx + 5xx)
        self.errors = DictPercentCounter(total=self.count)
        self._add_status_codes(
            pattern=DOWNLOADER_STATUS_CODES_ERRORS, target=self.errors)

    def _add_status_codes(self, pattern, target):
        for code, counter in self._get_response_codes(pattern).items():
            target.add_value(code, counter.count)

    def _get_response_codes(self, codes=None):
        codes = codes or ['[^/]+']
        return_codes = {}
        for code in codes:
            return_codes.update(self._get_response_code(code))
        return return_codes

    def _get_response_code(self, code):
        return dict(
            [
                (code, PercentCounter(count, self.count))
                for count, code in
                self._stats_analyzer.search(
                    pattern=DOWNLOADER_RESPONSE_STATUS + ('(%s)$' % code),
                    include_matches=True
                ).values()
            ]
        )


class SpiderMonitorMixin(StatsMonitorMixin, JobMonitorMixin):
    @property
    def crawler(self):
        if not self.data.crawler:
            raise NotConfigured('Crawler not available!')
        return self.data.crawler

    @property
    def spider(self):
        if not self.data.spider:
            raise NotConfigured('Spider not available!')
        return self.data.spider

    @property
    def responses(self):
        if not hasattr(self, '_responses'):
            self._responses = ResponsesInfo(self.stats)
        return self._responses



from __future__ import absolute_import
import os
import inspect

from spidermon.core.actions import ActionOptionsMetaclass, Action
from spidermon.templates import template_loader, Template
import six


class ActionWithTemplatesMetaclass(ActionOptionsMetaclass):
    def __new__(mcs, name, bases, attrs):
        cls = super(ActionWithTemplatesMetaclass, mcs).__new__(mcs, name, bases, attrs)
        mcs.add_class_templates(cls)
        return cls

    @classmethod
    def add_class_templates(mcs, cls):
        class_file = inspect.getfile(cls)
        class_path = os.path.dirname(class_file)
        template_loader.discover_folder(class_path)
        for path in cls.template_paths:
            template_path = os.path.join(class_path, path)
            template_loader.add_path(template_path)


class ActionWithTemplates(six.with_metaclass(ActionWithTemplatesMetaclass, Action)):
    template_paths = []
    context = None

    def __init__(self, context=None):
        super(ActionWithTemplates, self).__init__()
        self.context = context or self.context or {}

    def get_template(self, name):
        return template_loader.get_template(name)

    def render_text_template(self, template):
        template = Template(template)
        return template.render(self.get_template_context())

    def render_template(self, template):
        template = self.get_template(template)
        return template.render(self.get_template_context())

    def get_template_context(self):
        context = {
            'result': self.result,
            'data': self.data,
            'monitors_passed': self.monitors_passed,
            'monitors_failed': self.monitors_failed,
        }
        context.update(self.context)
        return context



from __future__ import absolute_import
from __future__ import print_function
import ast
import json
from slackclient import SlackClient

from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured
import six


class SlackMessageManager():
    sender_token = None
    sender_name = None

    def __init__(self, sender_token=None, sender_name=None):
        sender_token = sender_token or self.sender_token
        if not sender_token:
            raise NotConfigured('You must provide a slack token.')

        self.sender_name = sender_name or self.sender_name
        if not self.sender_name:
            raise NotConfigured('You must provide a slack sender name.')

        self._client = SlackClient(sender_token)
        self._users = None

    @property
    def users(self):
        if self._users is None:
            self._users = self._get_users_info()
        return self._users

    def send_message(self, to, text, parse=None, link_names=1, attachments=None, use_mention=False):
        if isinstance(to, list):
            return [self.send_message(
                to=recipient,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
                use_mention=use_mention,
            ) for recipient in to]
        elif to.startswith('@'):
                return self._send_user_message(
                    username=to,
                    text=text,
                    parse=parse,
                    link_names=link_names,
                    attachments=attachments)
        else:
            if use_mention:
                if to.startswith('#'):
                    text = '@channel: ' + text
                else:
                    text = '@group: ' + text
            return self._send_channel_message(
                channel=to,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
            )

    def _get_user_id(self, username):
        name = username[1:] if username.startswith('@') else username
        user = self.users.get(name, None)
        return user['id'] if user else None

    def _get_users_info(self):
        return dict([
            (member['name'].lower(), member)
            for member in self._api_call('users.list')['members']
        ])

    def _api_call(self, method, **kwargs):
        response = self._client.api_call(method, **kwargs)
        if isinstance(response, six.string_types):  # slackclient < v1.0
            response = json.loads(response)
        return response

    def _get_user_channel(self, user_id):
        return self._api_call('im.open', user=user_id)['channel']['id']

    def _send_user_message(self, username, text, parse='full', link_names=1, attachments=None):
        user_id = self._get_user_id(username)
        if user_id:
            user_channel = self._get_user_channel(user_id)
            return self._send_channel_message(
                channel=user_channel,
                text=text,
                parse=parse,
                link_names=link_names,
                attachments=attachments,
            )

    def _send_channel_message(self, channel, text, parse='full', link_names=1, attachments=None):
        return self._api_call(
            'chat.postMessage',
            channel=channel,
            text=text,
            parse=parse,
            link_names=link_names,
            attachments=self._parse_attachments(attachments),
            username=self.sender_name,
            icon_url=self.users[self.sender_name]['profile']['image_48'],
        )

    def _parse_attachments(self, attachments):
        if not attachments:
            return None
        else:
            python_attachments = ast.literal_eval(attachments)
            return json.dumps(python_attachments)


class SendSlackMessage(ActionWithTemplates):
    message = None
    attachments = None
    message_template = 'slack/default/message.jinja'
    attachments_template = 'slack/default/attachments.jinja'
    recipients = None
    sender_token = None
    sender_name = None
    include_message = True
    include_attachments = True
    fake = False

    def __init__(self,
                 sender_token=None, sender_name=None,
                 recipients=None,
                 message=None, message_template=None, include_message=None,
                 attachments=None, attachments_template=None, include_attachments=None,
                 fake=None):
        super(SendSlackMessage, self).__init__()
        self.fake = fake or self.fake
        if not self.fake:
            self.manager = SlackMessageManager(
                sender_token=sender_token or self.sender_token,
                sender_name=sender_name or self.sender_name,
            )
        else:
            self.manager = None
        self.recipients = recipients or self.recipients
        self.message = message or self.message
        self.message_template = message_template or self.message_template
        self.include_message = include_message or self.include_message
        self.attachments = attachments or self.attachments
        self.attachments_template = attachments_template or self.attachments_template
        self.include_attachments = include_attachments or self.include_attachments
        if not self.fake and not self.recipients:
            raise NotConfigured("You must provide at least one recipient for the message.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            'sender_token': crawler.settings.get('SPIDERMON_SLACK_SENDER_TOKEN'),
            'sender_name': crawler.settings.get('SPIDERMON_SLACK_SENDER_NAME'),
            'recipients': crawler.settings.get('SPIDERMON_SLACK_RECIPIENTS'),
            'message': crawler.settings.get('SPIDERMON_SLACK_MESSAGE'),
            'message_template': crawler.settings.get('SPIDERMON_SLACK_MESSAGE_TEMPLATE'),
            'attachments': crawler.settings.get('SPIDERMON_SLACK_ATTACHMENTS'),
            'attachments_template': crawler.settings.get('SPIDERMON_SLACK_ATTACHMENTS_TEMPLATE'),
            'include_message': crawler.settings.getbool('SPIDERMON_SLACK_INCLUDE_MESSAGE'),
            'include_attachments': crawler.settings.getbool('SPIDERMON_SLACK_INCLUDE_ATTACHMENTS'),
            'fake': crawler.settings.getbool('SPIDERMON_SLACK_FAKE'),
        }

    def run_action(self):
        message = self.get_message()
        attachments = self.get_attachments()
        if not self.fake:
            self.manager.send_message(
                to=self.recipients,
                text=message,
                attachments=attachments,
            )
        else:
            print('message:', message)
            print('attachments:', attachments)

    def get_message(self):
        if self.include_message:
            if self.message:
                return self.render_text_template(self.message)
            else:
                return self.render_template(self.message_template)
        else:
            return None

    def get_attachments(self):
        if self.include_attachments:
            if self.attachments:
                return self.render_text_template(self.attachments)
            else:
                return self.render_template(self.attachments_template)
        else:
            return None



from __future__ import absolute_import
from spidermon.contrib.monitors.mixins import (ValidationMonitorMixin,
                                               SpiderMonitorMixin)

from .factory import PythonExpressionsMonitor
from .context import Context


class ExpressionsMonitor(PythonExpressionsMonitor,
                         ValidationMonitorMixin,
                         SpiderMonitorMixin):
    def get_context_data(self):
        context = Context()
        attrs = [
            'stats',
            'crawler',
            'spider',
            'job',
            'validation',
            'responses'
        ]
        context.extend_via_attrs(self, attrs)
        return context




UNEXPECTED_FIELD = 'Unexpected field'
MISSING_REQUIRED_FIELD = 'Missing required field'
VALUE_NOT_IN_CHOICES = 'Value not in choices'
FIELD_TOO_LONG = 'Field too long'
FIELD_TOO_SHORT = 'Field too short'
INVALID_STRING = 'Invalid string'
REGEX_NOT_MATCHED = 'Regex not matched'
INVALID_DATETIME = 'Invalid datetime'
INVALID_DATE = 'Invalid date'
INVALID_INT = 'Invalid int'
INVALID_FLOAT = 'Invalid float'
INVALID_LONG = 'Invalid long'
INVALID_DECIMAL = 'Invalid decimal'
INVALID_BOOLEAN = 'Invalid boolean'
NUMBER_TOO_HIGH = 'Number too high'
NUMBER_TOO_LOW = 'Number too low'
INVALID_EMAIL = 'Invalid email'
INVALID_URL = 'Invalid URL'
INVALID_UUID = 'Invalid UUID'
INVALID_IPV4 = 'Invalid IPv4'
INVALID_IPV6 = 'Invalid IPv6'
INVALID_HOSTNAME = 'Invalid hostname'
INVALID_URI = 'Invalid URI'
INVALID_REGEX = 'Invalid regex'
INVALID_COLOR = 'Invalid color'
INVALID_HASH_LENGTH = 'Invalid hash length'
INVALID_HASH = 'Invalid hash'
LIST_TOO_SHORT = 'List too short'
LIST_TOO_LONG = 'List too long'
INVALID_LIST = 'Invalid list'
INVALID_DICT = 'Invalid dict'
INVALID_CHILD_CONTENT = 'Invalid child content'
INVALID_NUMBER = 'Invalid number'
INVALID_OBJECT = 'Invalid object'
INVALID_ARRAY = 'Invalid array'
NOT_NULL = 'Not null'
NOT_VALID_UNDER_ANY_SCHEMA = 'Not valid under any schema'
VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS = 'Valid for several exclusive schemas'
TOO_MANY_ITEMS = 'Too many items'
MISSING_DEPENDENT_FIELD = 'Missing dependent field'
TOO_MANY_PROPERTIES = 'Too many properties'
NOT_ENOUGH_PROPERTIES = 'Not enough properties'
NOT_MULTIPLE_OF = 'Not multiple of'
NOT_ALLOWED_VALUE = 'Not allowed value'
NOT_UNIQUE = 'Not unique'



from __future__ import absolute_import
import schematics
from schematics.models import Model
from schematics.types import (
    StringType,
    DateTimeType, DateType,
    FloatType, IntType, LongType, DecimalType,
    BooleanType, EmailType, URLType,
    UUIDType, IPv4Type, MD5Type, SHA1Type,
)
from schematics.types.compound import ListType, DictType, ModelType

from spidermon.contrib.validation import SchematicsValidator, messages


SCHEMATICS1 = schematics.__version__.startswith('1.')


def test_rogue_fields():
    """
    messages:
        - UNEXPECTED_FIELD
    """
    _test_data(
        model=Model,
        data={'a': 1},
        expected=(False, {'a': [messages.UNEXPECTED_FIELD]}),
    )
    _test_data(
        model=Model,
        data={'a': 1},
        expected=(True, {}),
        strict=False
    )


def test_required():
    """
    messages:
        - MISSING_REQUIRED_FIELD
    """
    class DataRequired(Model):
        a = StringType(required=True)

    class DataNotRequired(Model):
        a = StringType(required=False)

    _test_data(
        model=DataRequired,
        data={},
        expected=(False, {'a': [messages.MISSING_REQUIRED_FIELD]}),
    )
    _test_data(
        model=DataNotRequired,
        data={},
        expected=(True, {}),
    )


def test_choices():
    """
    messages:
        - VALUE_NOT_IN_CHOICES
    """
    class Data(Model):
        a = StringType(choices=['a', 'b'])
        b = IntType(choices=[1, 2, 3])

    _test_data(
        model=Data,
        data={},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={
            'a': 'b',
            'b': 3,
        },
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={
            'a': 'c',
            'b': 4,
        },
        expected=(False, {
            'a': [messages.VALUE_NOT_IN_CHOICES],
            'b': [messages.VALUE_NOT_IN_CHOICES],
        }),
    )


def test_string_valid():
    """
    messages:
        - INVALID_STRING
    """
    class Data(Model):
        a = StringType()

    _test_data(
        model=Data,
        data={'a': 'hello there!'},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': []},
        expected=(False, {'a': [messages.INVALID_STRING]}),
    )


def test_string_lengths():
    """
    messages:
        - FIELD_TOO_SHORT
        - FIELD_TOO_LONG
    """
    class Data(Model):
        a = StringType(min_length=2, max_length=5)

    _test_data(
        model=Data,
        data={'a': '12'},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': '12345'},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': '1'},
        expected=(False, {'a': [messages.FIELD_TOO_SHORT]}),
    )
    _test_data(
        model=Data,
        data={'a': '123456'},
        expected=(False, {'a': [messages.FIELD_TOO_LONG]}),
    )


def test_string_regex():
    """
    messages:
        - REGEX_NOT_MATCHED
    """
    class Data(Model):
        a = StringType(regex='.*def.*')

    _test_data(
        model=Data,
        data={'a': 'abc'},
        expected=(False, {'a': [messages.REGEX_NOT_MATCHED]}),
    )
    _test_data(
        model=Data,
        data={'a': 'abcdefghi'},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': 'def'},
        expected=(True, {}),
    )


def test_datetime():
    """
    messages:
        - INVALID_DATETIME
    """
    class Data(Model):
        a = DateTimeType()

    class DataWithFormats(Model):
        a = DateTimeType(formats=('%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S'))

    if SCHEMATICS1:
        INVALID = [
            '2015-05-13 13:35:15.718978',
            '2015-05-13 13:35:15',
        ]
    else:
        INVALID = [
            'foo',
            '1-2-3',
        ]
        CUSTOM_FORMAT = [
            '2015-05-13 13:35:15.718978',
            '2015-05-13 13:35:15',
        ]
    VALID = [
        '2015-05-13T13:35:15.718978',
        '2015-05-13T13:35:15',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DATETIME,
    )
    if SCHEMATICS1:
        for dt in INVALID:
            _test_data(
                model=DataWithFormats,
                data={'a': dt},
                expected=(True, {}),
            )
    else:
        for dt in CUSTOM_FORMAT:
            _test_data(
                model=DataWithFormats,
                data={'a': dt},
                expected=(True, {}),
            )


def test_date():
    """
    messages:
        - INVALID_DATE
    """
    class Data(Model):
        a = DateType()

    INVALID = [
        '2015-05-13 13:35:15',
        '13-05-2013',
        '2015-20-13',
        '2015-01-40',
    ]
    VALID = [
        '2015-05-13',
        '2050-01-01',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DATE,
    )


def test_int():
    """
    messages:
        - INVALID_INT
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = IntType(min_value=-10, max_value=10)
        b = IntType()

    INVALID = [
        '',
        'a',
        '2a',
        '2015-05-13 13:35:15',
        '7.2',
    ]
    VALID = [
        '1',
        '8',
        '-2',
        '-7',
        1,
        8,
        -2,
        -7,
    ]
    if SCHEMATICS1:
        VALID.append(7.2)
    else:
        INVALID.append(7.2)
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_INT,
    )
    _test_data(
        model=Data,
        data={'a': -20},
        expected=(False, {'a': [messages.NUMBER_TOO_LOW]}),
    )
    _test_data(
        model=Data,
        data={'a': 11},
        expected=(False, {'a': [messages.NUMBER_TOO_HIGH]}),
    )


def test_float():
    """
    messages:
        - INVALID_FLOAT
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = FloatType(min_value=-10, max_value=10)

    INVALID = [
        '',
        'a',
        '2a',
        '2015-05-13 13:35:15',
    ]
    VALID = [
        '1',
        '-2',
        '8',
        '2.3',
        '5.2354958',
        '-9.231',
        1,
        -2,
        8,
        2.3,
        5.2354958,
        -9.231,
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_FLOAT,
    )
    _test_data(
        model=Data,
        data={'a': -20},
        expected=(False, {'a': [messages.NUMBER_TOO_LOW]}),
    )
    _test_data(
        model=Data,
        data={'a': 11},
        expected=(False, {'a': [messages.NUMBER_TOO_HIGH]}),
    )


def test_long():
    """
    messages:
        - INVALID_LONG
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = LongType(min_value=-10, max_value=10)

    INVALID = [
        '',
        'a',
        '2a',
        '2015-05-13 13:35:15',
        '2.3',
        '5.2354958',
    ]
    VALID = [
        '1',
        '-2',
        '8',
        1,
        -2,
        8,
    ]
    if SCHEMATICS1:
        expected_error = messages.INVALID_LONG
    else:
        expected_error = messages.INVALID_INT
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=expected_error,
    )
    _test_data(
        model=Data,
        data={'a': -20},
        expected=(False, {'a': [messages.NUMBER_TOO_LOW]}),
    )
    _test_data(
        model=Data,
        data={'a': 11},
        expected=(False, {'a': [messages.NUMBER_TOO_HIGH]}),
    )


def test_decimal():
    """
    messages:
        - INVALID_DECIMAL
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = DecimalType(min_value=-10, max_value=10)

    INVALID = [
        '',
        'a',
        '2a',
        '2015-05-13 13:35:15',
    ]
    VALID = [
        '1',
        '-2',
        '8',
        '2.3',
        '5.2354958',
        '-9.231',
        1,
        -2,
        8,
        2.3,
        5.2354958,
        -9.231,
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DECIMAL,
    )
    _test_data(
        model=Data,
        data={'a': -20},
        expected=(False, {'a': [messages.NUMBER_TOO_LOW]}),
    )
    _test_data(
        model=Data,
        data={'a': 11},
        expected=(False, {'a': [messages.NUMBER_TOO_HIGH]}),
    )


def test_boolean():
    """
    messages:
        - INVALID_BOOLEAN
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = BooleanType()

    INVALID = [
        '',
        'a',
        '2'
        'TRUE',
        'FALSE',
        'TruE',
        'FalsE',
    ]
    VALID = [
        0,
        1,
        '0',
        '1',
        'True',
        'False',
        'true',
        'false',
        True,
        False
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_BOOLEAN,
    )


def test_email():
    """
    messages:
        - INVALID_EMAIL
        - NUMBER_TOO_LOW
        - NUMBER_TOO_HIGH
    """
    class Data(Model):
        a = EmailType()

    INVALID = [
        '',
        'johndoe',
        'johndoe@domain'
        'johndoe@domain.'
        '@domain'
        '@domain.com'
        'domain.com'
    ]
    VALID = [
        'johndoe@domain.com',
        'john.doe@domain.com',
        'john.doe@sub.domain.com',
        'j@sub.domain.com',
        'j@d.com',
        'j@domain.co.uk',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_EMAIL,
    )


def test_url():
    """
    messages:
        - INVALID_URL
    """
    class Data(Model):
        a = URLType()

    INVALID = [
        '',
        'http://',
        'http://www.',
        'www.',
        'http://www. .com',
        'domain.com',
        'www.domain.com',
        'http:/www.domain.com',
        'http//www.domain.com',
        'http:www.domain.com',
        'htp://domain.com/',
        'http://sub.domain.com\\en-us\\default.aspx\\',
        'http:\\\\msdn.domain.com\\en-us\\library\\default.aspx\\',
        'http:\\\\www.domain.com\\leafnode-L1.html',
        './',
        '../',
        'http:\\\\www.domain.com\\leafnode-L1.xhtml\\',
    ]
    VALID = [
        'http://www.domain',
        'http://www.com',
        'http://www.domain.com.',
        'http://www.domain.com/.',
        'http://www.domain.com/..',
        'http://www.domain.com//cataglog//index.html',
        'http://www.domain.net/',
        'http://www.domain.com/level2/leafnode-L2.xhtml/',
        'http://www.domain.com/level2/level3/leafnode-L3.xhtml/',
        'http://www.domain.com?pageid=123&testid=1524',
        'http://www.domain.com/do.html#A',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_URL,
    )


def test_uuid():
    """
    messages:
        - INVALID_UUID
    """
    class Data(Model):
        a = UUIDType()

    INVALID = [
        '',
        '678as6sd88ads67',
        '678as6sd88ads67-alskjlasd',
        'xxxxxxxx-xxxx-Mxxx-Nxxx-xxxxxxxxxxxx',
        '2.25.290383009913173870543740933812899923227',
    ]
    VALID = [
        '12345678-1234-5678-1234-567812345678',
        '12345678123456781234567812345678',
        'urn:uuid:12345678-1234-5678-1234-567812345678',
        'cfc63f3f-f3a7-465a-8183-acf055c6d472',
        '00000000-0000-0000-0000-000000000000',
        '01234567-89ab-cdef-0123456789abcdef',
        '0123456789abcdef0123456789abcdef',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_UUID,
    )


def test_ipv4type():
    """
    messages:
        - INVALID_IPV4
    """
    class Data(Model):
        a = IPv4Type()

    INVALID = [
        '',
        '0',
        '0.',
        '0.0',
        '0.0.',
        '0.0.0',
        '0.0.0.0.',
        '0.0.0.0.0',
        '256.256.256.256',
        '2002:4559:1FE2::4559:1FE2',
        '2002:4559:1FE2:0:0:0:4559:1FE2',
        '2002:4559:1FE2:0000:0000:0000:4559:1FE2',
    ]
    VALID = [
        '98.139.180.149',
        '69.89.31.226',
        '192.168.1.1',
        '127.0.0.0',
        '0.0.0.0',
        '255.255.255.255',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_IPV4,
    )


def test_md5():
    """
    messages:
        - INVALID_HASH
        - INVALID_HASH_LENGTH
    """
    class Data(Model):
        a = MD5Type()

    INVALID = [
        '_b1a9953c4611296a827abf8a47804d7',
        'zb1a9953c4611296a827abf8a47804d7',
        'Gb1a9953c4611296a827abf8a47804d7',
        # FIXME: PY3: schematics uses integer conversion for validating hex and
        # Py3 integers can contain underscores.
        # '8b1_9953c4611296a827abf8c47804d1',
    ]
    VALID = [
        '8b1a9953c4611296a827abf8c47804d7',
        '7dd4bbe8a38600b556f79ca44c9b5132',
        '11111111111111111111111111111111',
        '8B1A9953C4611296A827ABF8C47804D1',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_HASH,
    )
    _test_data(
        model=Data,
        data={'a': '8b1a9953c4611296a827abf8c47804d'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={'a': '8b1a9953c4611296'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={'a': '8b1a9953c46112968b1a9953c46112968b1a9953c4611296'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )


def test_sha1():
    """
    messages:
        - INVALID_HASH
        - INVALID_HASH_LENGTH
    """
    class Data(Model):
        a = SHA1Type()

    INVALID = [
        '_03d40e1a2ede7e31f3c3b45a9e87d12ed33402e',
        'g03d40e1a2ede7e31f3c3b45a9e87d12ed33402e',
        'z03d40e1a2ede7e31f3c3b45a9e87d12ed33402e',
        'G03d40e1a2ede7e31f3c3b45a9e87d12ed33402e',
        # FIXME: PY3: schematics uses integer conversion for validating hex and
        # Py3 integers can contain underscores.
        # 'a03d_0e1a2ede7e31f3c3b45a9e87d12ed33402e',
    ]
    VALID = [
        'a03d70e1a2ede7e31f3c3b45a9e87d12ed33402e',
    ]
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_HASH,
    )
    _test_data(
        model=Data,
        data={'a': 'a03d70e1a2ede7e31f3c3b45a9e87d12ed33402'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={'a': '8b1a9953c4611296'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )
    _test_data(
        model=Data,
        data={'a': '8b1a9953c46112968b1a9953c46112968b1a9953c4611296'},
        expected=(False, {'a': [messages.INVALID_HASH_LENGTH]}),
    )


def test_list():
    """
    messages:
        - INVALID_LIST
        - LIST_TOO_SHORT
        - LIST_TOO_LARGE
        - INVALID_INT
    """
    class Data(Model):
        a = ListType(field=IntType(), min_size=3, max_size=5)

    _test_data(
        model=Data,
        data={'a': [1, 2, 3]},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': ["1", "2", "3"]},
        expected=(True, {}),
    )
    _test_data(
        model=Data,
        data={'a': Data},
        expected=(False, {'a': [messages.INVALID_LIST]}),
    )
    if SCHEMATICS1:
        _test_data(
            model=Data,
            data={'a': ["a", "b", "c"]},
            expected=(False, {'a': [messages.INVALID_INT]}),
        )
    else:
        _test_data(
            model=Data,
            data={'a': ["a", "b", "c"]},
            expected=(
                False,
                {
                    'a.0': [messages.INVALID_INT],
                    'a.1': [messages.INVALID_INT],
                    'a.2': [messages.INVALID_INT]
                }
            ),
        )
    _test_data(
        model=Data,
        data={'a': [1, 2]},
        expected=(False, {'a': [messages.LIST_TOO_SHORT]}),
    )
    _test_data(
        model=Data,
        data={'a': [1, 2, 3, 4, 5, 6]},
        expected=(False, {'a': [messages.LIST_TOO_LONG]}),
    )


def test_dict():
    """
    messages:
        - INVALID_DICT
        - INVALID_INT
    """
    class Data(Model):
        a = DictType(field=IntType)

    INVALID = [
        'a',
        Data,
    ]
    VALID = [
        {},
        {'some': 1},
    ]
    if SCHEMATICS1:
        VALID.append([])
    else:
        INVALID.append([])
    _test_valid_invalid(
        model=Data,
        valid=VALID,
        invalid=INVALID,
        expected_error=messages.INVALID_DICT,
    )
    if SCHEMATICS1:
        _test_data(
            model=Data,
            data={'a': {'some': "a"}},
            expected=(False, {'a': [messages.INVALID_INT]}),
        )
    else:
        _test_data(
            model=Data,
            data={'a': {'some': "a"}},
            expected=(False, {'a.some': [messages.INVALID_INT]}),
        )


def test_models():
    """
    messages:
        - UNEXPECTED_FIELD
        - MISSING_REQUIRED_FIELD
        - INVALID_FLOAT
    """
    class Coordinates(Model):
        latitude = FloatType(required=True)
        longitude = FloatType(required=True)

    class Geo(Model):
        coordinates = ModelType(Coordinates, required=True)

    class Data(Model):
        geo = ModelType(Geo, required=True)

    _test_data(
        model=Data,
        data={'a': {}},
        expected=(False, {
            'a': [messages.UNEXPECTED_FIELD],
            'geo': [messages.MISSING_REQUIRED_FIELD],
        }),
    )
    _test_data(
        model=Data,
        data={'geo': None},
        expected=(False, {
            'geo': [messages.MISSING_REQUIRED_FIELD]
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {}},
        expected=(False, {
            'geo.coordinates': [messages.MISSING_REQUIRED_FIELD]
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {
            'coordinates': None,
        }},
        expected=(False, {
            'geo.coordinates': [messages.MISSING_REQUIRED_FIELD]
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {
            'coordinates': {},
        }},
        expected=(False, {
            'geo.coordinates.latitude': [messages.MISSING_REQUIRED_FIELD],
            'geo.coordinates.longitude': [messages.MISSING_REQUIRED_FIELD],
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {
            'coordinates': {
                'latitude': None,
                'longitude': None,
            },
        }},
        expected=(False, {
            'geo.coordinates.latitude': [messages.MISSING_REQUIRED_FIELD],
            'geo.coordinates.longitude': [messages.MISSING_REQUIRED_FIELD],
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {
            'coordinates': {
                'latitude': "y",
                'longitude': "x",
            },
        }},
        expected=(False, {
            'geo.coordinates.latitude': [messages.INVALID_FLOAT],
            'geo.coordinates.longitude': [messages.INVALID_FLOAT],
        }),
    )
    _test_data(
        model=Data,
        data={'geo': {
            'coordinates': {
                'latitude': 40.42,
                'longitude': -3.71,
            },
        }},
        expected=(True, {}),
    )


def test_multiple_errors_per_field():
    """
    messages:
        - FIELD_TOO_SHORT
        - REGEX_NOT_MATCHED
    """
    class Data(Model):
        a = StringType(min_length=3, regex=r'foo')

    data = {'a': 'z'}
    v = SchematicsValidator(Data)
    result = v.validate(data, strict=True)
    assert result[0] is False
    error_messages = result[1]
    assert 'a' in error_messages
    expected = [messages.FIELD_TOO_SHORT, messages.REGEX_NOT_MATCHED]
    assert sorted(error_messages['a']) == expected


def _test_data(model, data, expected, strict=True):
    v = SchematicsValidator(model)
    assert expected == v.validate(data, strict=strict)


def _test_valid_invalid(model, valid, invalid, expected_error, expected_field='a'):
    for dt in valid:
        _test_data(
            model=model,
            data={expected_field: dt},
            expected=(True, {}),
        )
    for dt in invalid:
        _test_data(
            model=model,
            data={expected_field: dt},
            expected=(False, {expected_field: [expected_error]}),
        )



from __future__ import absolute_import
import pytest

from spidermon import MonitorSuite
from spidermon.exceptions import (InvalidMonitor, InvalidMonitorIterable,
                                  InvalidMonitorClass, InvalidMonitorTuple,
                                  NotAllowedMethod)

from .fixtures.suites import *


class SuiteDefinition(object):
    def __init__(self, suite_class, monitors=None, expected_number_of_monitors=0):
        self.suite_class = suite_class
        self.monitors = monitors or []
        self.expected_number_of_monitors = expected_number_of_monitors


CLASS_SUITE_DEFINITIONS = [
    # ---------------------------------------------------------------------------------
    #               suite_class    monitors               expected_number_of_monitors
    # ---------------------------------------------------------------------------------

    # no monitors
    SuiteDefinition(EmptySuite,    [],                                              0),
    SuiteDefinition(Suite01,       [],                                              3),
    SuiteDefinition(Suite02,       [],                                              5),
    SuiteDefinition(Suite03,       [],                                              8),
    SuiteDefinition(Suite04,       [],                                             13),

    # monitors as classes
    SuiteDefinition(EmptySuite,    [EmptyMonitor],                                  0),
    SuiteDefinition(EmptySuite,    [Monitor01],                                     3),
    SuiteDefinition(EmptySuite,    [Monitor02],                                     2),
    SuiteDefinition(EmptySuite,    [Monitor01, Monitor02],                          5),
    SuiteDefinition(EmptySuite,    [Suite01],                                       3),
    SuiteDefinition(EmptySuite,    [Suite01, Suite02],                              8),
    SuiteDefinition(EmptySuite,    [Suite01, Suite02, Monitor01, Monitor02],       13),
    SuiteDefinition(Suite01,       [Suite01, Suite02, Monitor01, Monitor02],       16),
    SuiteDefinition(Suite02,       [Suite01, Suite02, Monitor01, Monitor02],       18),
    SuiteDefinition(Suite03,       [Suite01, Suite02, Monitor01, Monitor02],       21),
    SuiteDefinition(Suite04,       [Suite01, Suite02, Monitor01, Monitor02],       26),

    # monitors as instances
    SuiteDefinition(EmptySuite,    [Suite01()],                                     3),
    SuiteDefinition(EmptySuite,    [Suite02()],                                     5),
    SuiteDefinition(EmptySuite,    [Suite03()],                                     8),
    SuiteDefinition(EmptySuite,    [Suite04()],                                    13),
    SuiteDefinition(EmptySuite,    [Suite01(), Suite02(), Suite03(), Suite04()],   29),

    # monitors as tuples
    SuiteDefinition(EmptySuite,    [('M1', Monitor01)],                             3),
    SuiteDefinition(EmptySuite,    [('M1', Monitor01), ('M2', Monitor02)],          5),
    SuiteDefinition(EmptySuite,    [('S', Suite04)],                               13),
]
INVALID_SUITE_DEFINITION_NON_ITERABLE   = SuiteDefinition(EmptySuite, 10)
INVALID_SUITE_DEFINITION_NONE           = SuiteDefinition(EmptySuite, [None])
INVALID_SUITE_DEFINITION_NUMBER         = SuiteDefinition(EmptySuite, [10])
INVALID_SUITE_DEFINITION_CLASS          = SuiteDefinition(EmptySuite, [object])
INVALID_SUITE_DEFINITION_TUPLE1         = SuiteDefinition(EmptySuite, [(1,)])
INVALID_SUITE_DEFINITION_TUPLE3         = SuiteDefinition(EmptySuite, [(1, 2, 3)])
INVALID_SUITE_DEFINITION_TUPLE2_WRONG1  = SuiteDefinition(EmptySuite, [(1, 2)])
INVALID_SUITE_DEFINITION_TUPLE2_WRONG2  = SuiteDefinition(EmptySuite, [('A', 2)])

INVALID_SUITE_DEFINITIONS = [
    # --------------------------------------------------------------
    # definition                                raises
    # --------------------------------------------------------------
    (INVALID_SUITE_DEFINITION_NONE,             InvalidMonitor),
    (INVALID_SUITE_DEFINITION_NUMBER,           InvalidMonitor),
    (INVALID_SUITE_DEFINITION_CLASS,            InvalidMonitorClass),
    (INVALID_SUITE_DEFINITION_TUPLE1,           InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE3,           InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE2_WRONG1,    InvalidMonitorTuple),
    (INVALID_SUITE_DEFINITION_TUPLE2_WRONG2,    InvalidMonitor),
]


def test_creation_error_non_iterable():
    with pytest.raises(InvalidMonitorIterable):
        _test_creation_from_init(INVALID_SUITE_DEFINITION_NON_ITERABLE)
    with pytest.raises(InvalidMonitorIterable):
        _test_creation_from_add_monitors(INVALID_SUITE_DEFINITION_NON_ITERABLE)


def test_creation_error_invalid_from_init():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_init(definition)


def test_creation_error_invalid_from_add_monitors():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_add_monitors(definition)


def test_creation_error_invalid_from_add_monitor():
    for definition, exception_to_raise in INVALID_SUITE_DEFINITIONS:
        with pytest.raises(exception_to_raise):
            _test_creation_from_add_monitor(definition)


def test_creation_from_init():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_init(definition)


def test_creation_from_add_monitors():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_add_monitors(definition)


def test_creation_from_add_monitor():
    for definition in CLASS_SUITE_DEFINITIONS:
        _test_creation_from_add_monitor(definition)


def _test_creation_from_init(definition):
    suite = definition.suite_class(monitors=definition.monitors)
    check_suite(
        suite=suite,
        expected_number_of_monitors=definition.expected_number_of_monitors,
    )


def test_not_allowed_methods():
    suite = EmptySuite()
    with pytest.raises(NotAllowedMethod):
        suite.addTest()
    with pytest.raises(NotAllowedMethod):
        suite.addTests()


def _test_creation_from_add_monitors(definition):
    suite = definition.suite_class()
    suite.add_monitors(definition.monitors)
    check_suite(
        suite=suite,
        expected_number_of_monitors=definition.expected_number_of_monitors,
    )


def _test_creation_from_add_monitor(definition):
    suite = definition.suite_class()
    for monitor in definition.monitors:
        suite.add_monitor(monitor)
    check_suite(
        suite=suite,
        expected_number_of_monitors=definition.expected_number_of_monitors,
    )


def check_suite(suite, expected_number_of_monitors):
    #print
    #print suite.debug_tree()
    #print
    assert isinstance(suite, MonitorSuite)
    assert suite.number_of_monitors == expected_number_of_monitors
    for monitor in suite:
        assert isinstance(monitor, (Monitor, MonitorSuite))
    all_monitors = suite.all_monitors
    assert len(all_monitors) == expected_number_of_monitors
    for test in all_monitors:
        assert isinstance(test, (Monitor, Monitor))



from six import PY2

SYNTAXERROR_EXPRESSIONS = [
    '?',
    'a string',
    'a=',
]

INVALID_EXPRESSIONS = [
    None,
    0,
    '',
    """
    """,
    'a > b\na > b',
    'import os',
    'from os import *',
    'for i in range(10): pass',
    'def something(): pass',
    'lambda x: x',
    "del a",
    # "print(a)",
    "raise Exception",
    "pass",
    "return",
    "yield x",

    # Assignment Operations
    "a = 1",
    "a += 1",
    "a -= 1",
    "a *= 1",
    "a /= 1",
    "a %= 1",
    "a **= 1",
    "a //= 1",
]

VALID_EXPRESSIONS = [
    # Strings
    "'single quoted string'",
    "\"double quoted string\"",
    "u'unicode single quoted string'",
    "u\"unicode double quoted string\"",

    # Numbers
    "0",
    ".5",
    "-10",
    "10",
    "1.0",
    "-1.0",
    "3.14j",

    # Sequences
    "[1, 2, 3]",
    "{'a': 1, 'b': 1}",
    "set()",
    "(1, 2)",
    "len(a)",

    # Constants
    "None",
    "True",
    "False",

    # Arithmetic Operations
    "-a",
    "a + 1",
    "a - 1",
    "a / 1",
    "a * 1",
    "a % 1",
    "a ** 1",
    "a // 1",

    # Comparison Operations
    "a == b",
    "a != b",
    "a > b",
    "a < b",
    "a >= b",
    "a <= b",
    "1 < a < 10",

    # Bitwise Operations
    "a & b",
    "a | b",
    "a ^ b",
    "~a",
    "a << b",
    "a >> b",

    # Logical Operations
    "a and b",
    "a or b",
    "not a",

    # Membership Operations
    "a in b",
    "a not in b",

    # Identity Operations
    "a is 10",
    "a is not 10",

    # Inline if statement
    "a if b else c",

    # Subscripting
    "a[1]",
    "a[1:2]",
    "a[1:2, 3]",

    # Comprehensions
    "[i for i in range(10)]",
    "{i: i**2 for i in range(10)}",
    "{i for i in range(10)}",
    "[n for n in range(10) if n>5]",

    # Attribute access
    "stats.scraped_items"
]


EXPRESSIONS_TO_EVALUATE = [
    ('stats.item_scraped_count == 10000', False),
    ('stats.item_scraped_count != 10000', True),
    ('stats.item_scraped_count < 10000', False),
    ('stats.item_scraped_count > 10000', True),
    ('stats.item_scraped_count <= 10000', False),
    ('stats.item_scraped_count >= 10000', True),
    ('10000 < stats.item_scraped_count < 100000', True),
    ('stats.item_scraped_count == .5', False),
    ('stats.item_scraped_count < .5', False),
    ('stats.item_scraped_count > .5', True),

    ('stats.finish_reason in ["finished"]', True),
    ('stats.finish_reason in ["a", "b"]', False),
    ('stats.finish_reason not in ["a", "b"]', True),
    ('stats.finish_reason in {"finished": 0}', True),
    ('stats.finish_reason in {"finished": 0}.keys()', True),
    ('stats.finish_reason in ("finished",)', True),
    ('len(stats.finish_reason) == len("finished")', True),
    ('stats.finish_reason is not None', True),
    ('len(stats.finish_reason) > 0', True),
    ('(len(stats.finish_reason) > 0) == True', True),
    ('(len(stats.finish_reason) > 0) != False', True),
    ('(len(stats.finish_reason) > 0) is not False', True),
    ('(len(stats.finish_reason) > 0) is True', True),

    ('stats["downloader/exception_count"] == 16', True),
    ('-stats["downloader/exception_count"] == -16', True),
    ('stats["downloader/exception_count"]+1 == 17', True),
    ('stats["downloader/exception_count"]-1 == 15', True),
    ('stats["downloader/exception_count"]/2 == 8', True),
    ('stats["downloader/exception_count"]*2 == 32', True),
    ('stats["downloader/exception_count"]%3 == 1', True),
    ('stats["downloader/exception_count"]**2 == 256', True),
    ('stats["downloader/exception_count"]//2 == 8', True),
    ('stats["downloader/exception_count"]*2/2 == 16', True),

    ('stats.has_errors and stats.has_redirections', True),
    ('stats.has_errors or stats.has_redirections', True),
    ('not stats.has_errors', False),
    ('stats.has_errors is True', True),
    ('stats.has_errors is not True', False),
    ('True if stats.has_errors else False', True),
    ('True if not stats.has_errors else False', False),
    ('stats.item_scraped_count in range(29830, 29840)', True),
]


if PY2:
    VALID_EXPRESSIONS.extend([
        "51924361L",
        "a <> b",
    ])
    EXPRESSIONS_TO_EVALUATE.extend([
        ('stats.item_scraped_count <> 10000', True),
    ])



from __future__ import absolute_import
from spidermon import MonitorSuite


from .cases import *


class EmptySuite(MonitorSuite):
    pass


class Suite01(MonitorSuite):
    monitors = [
        Monitor01,
    ]


class Suite02(MonitorSuite):
    monitors = [
        Suite01,
        Monitor02,
    ]


class Suite03(MonitorSuite):
    monitors = [
        Suite01,
        Suite02,
    ]


class Suite04(MonitorSuite):
    monitors = [
        Suite01,
        Suite02,
        Monitor01,
        Monitor02,
    ]



from __future__ import absolute_import
import inspect
import os
import datetime
import pprint as pretty_print

from jinja2 import Environment, FileSystemLoader, Template

DEFAULT_TEMPLATE_FOLDERS = ['templates']


def get_log_errors(logs):
    return [e for e in logs.list() if e['level'] >= 40]


def make_list(obj):
    return list(obj)


def pprint(obj):
    return pretty_print.pformat(obj)


def format_time(time):
    if not isinstance(time, datetime.timedelta):
        time = datetime.timedelta(seconds=int(time / 1000.))
    return ':'.join(str(time).split(':')[:2])+'h'


FILTERS = {
    'pprint': pprint,
    'list': make_list,
    'get_log_errors': get_log_errors,
    'format_time': format_time,
}
GLOBALS = {
    'datetime': datetime,
    'str': str,
}


class TemplateLoader(object):
    def __init__(self):
        self.paths = []
        self.reload_env()

    def add_path(self, path):
        if path not in self.paths and os.path.isdir(path):
            self.paths.append(path)
            self.reload_env()

    def auto_discover(self, path=None, folder=None):
        caller_folder = os.path.dirname(inspect.stack()[1][1])
        if path:
            caller_folder = os.path.join(caller_folder, path)
        if folder:
            self.add_path(os.path.join(caller_folder, folder))
        else:
            self.discover_folder(caller_folder)

    def discover_folder(self, candidate_folder):
        for folder in [os.path.join(candidate_folder, dir)
                       for dir in DEFAULT_TEMPLATE_FOLDERS]:
            self.add_path(folder)

    def reload_env(self):
        loader = FileSystemLoader(self.paths)
        self.env = Environment(
            loader=loader,
            lstrip_blocks=True,
            trim_blocks=True,
            )
        for filter_name, filter in FILTERS.items():
            self.env.filters[filter_name] = filter

        for global_name, global_value in GLOBALS.items():
            self.env.globals[global_name] = global_value

    def get_template(self, name):
        return self.env.get_template(name)

template_loader = TemplateLoader()



STATS_A = {
    'item_scraped_count': 150,
}

STATS_B = {
    'item_scraped_count': 0,
}

STATS_EMPTY = {}

STATS_TO_EVALUATE = {
    'downloader/exception_count': 16,
    'downloader/response_count': 63785,
    'finish_reason': 'finished',
    'item_scraped_count': 29836,
    'has_errors': True,
    'has_redirections': True,
}




from __future__ import absolute_import
import re


class MessageTranslator(object):
    messages = {}

    def __init__(self):
        self.compiled_messages = dict([(m, re.compile(m)) for m in self.messages])

    def translate_messages(self, messages):
        return [self.translate_message(m) for m in messages]

    def translate_message(self, message):
        for target_message, pattern in self.compiled_messages.items():
            if pattern.search(message):
                return self.messages[target_message]  # TO-DO: Add substitution?
        return message




from __future__ import absolute_import
from jsonschema._format import FormatChecker, _draft_checkers
from jsonschema.compat import str_types

from spidermon.contrib.validation.utils import is_valid_url, is_valid_email
from six import iterkeys


def is_url(instance):
    if not isinstance(instance, str_types):
        return True
    return is_valid_url(instance)


def is_email(instance):
    if not isinstance(instance, str_types):
        return True
    return is_valid_email(instance)


_spidermon_checkers = {
    'url': (is_url, ()),
    'email': (is_email, ()),
}

#_draft4_checkers = ['email', 'ipv4', 'ipv6', 'hostname', 'uri', 'date-time', 'regex']

for format_name, (func, raises) in _spidermon_checkers.items():
    FormatChecker.cls_checks(format_name, raises)(func)


format_checker = FormatChecker(_draft_checkers["draft4"] + list(iterkeys(_spidermon_checkers)))



from __future__ import absolute_import
from __future__ import print_function
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured

from premailer import transform


class SendEmail(ActionWithTemplates):
    sender = None
    subject = None
    subject_template = None
    to = None
    cc = None
    bcc = None
    reply_to = None
    body_text = None
    body_text_template = None
    body_html = None
    body_html_template = None
    fake = False

    def __init__(self,
                 sender,
                 to, cc=None, bcc=None, reply_to=None,
                 subject=None, subject_template=None,
                 body_text=None, body_text_template=None,
                 body_html=None, body_html_template=None,
                 fake=None,
                 *args, **kwargs):
        super(SendEmail, self).__init__(*args, **kwargs)
        self.sender = sender or self.sender
        self.subject = subject or self.subject
        self.subject_template = subject_template or self.subject_template
        self.to = to or self.to
        self.cc = cc or self.cc
        self.bcc = bcc or self.bcc
        self.reply_to = reply_to or self.reply_to
        self.body_text = body_text or self.body_text
        self.body_text_template = body_text_template or self.body_text_template
        self.body_html = body_html or self.body_html
        self.body_html_template = body_html_template or self.body_html_template
        self.fake = fake or self.fake
        if not self.fake and not self.to:
            raise NotConfigured("You must provide at least one recipient for the message.")
        if not self.subject:
            raise NotConfigured("You must provide a subject for the message.")
        if not (self.body_text or body_text_template or body_html or self.body_html_template):
            raise NotConfigured("You must provide a body for the message.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            'sender': crawler.settings.get('SPIDERMON_EMAIL_SENDER'),
            'subject': crawler.settings.get('SPIDERMON_EMAIL_SUBJECT'),
            'subject_template': crawler.settings.get('SPIDERMON_EMAIL_SUBJECT_TEMPLATE'),
            'to': crawler.settings.get('SPIDERMON_EMAIL_TO'),
            'cc': crawler.settings.get('SPIDERMON_EMAIL_CC'),
            'bcc': crawler.settings.get('SPIDERMON_EMAIL_BCC'),
            'reply_to': crawler.settings.get('SPIDERMON_EMAIL_REPLY_TO'),
            'body_text': crawler.settings.get('SPIDERMON_BODY_TEXT'),
            'body_text_template': crawler.settings.get('SPIDERMON_BODY_TEXT_TEMPLATE'),
            'body_html': crawler.settings.get('SPIDERMON_BODY_HTML'),
            'body_html_template': crawler.settings.get('SPIDERMON_BODY_HTML_TEMPLATE'),
            'fake': crawler.settings.getbool('SPIDERMON_EMAIL_FAKE'),
            'context': crawler.settings.getdict('SPIDERMON_EMAIL_CONTEXT'),
        }

    def run_action(self):
        message = self.get_message()
        if not self.fake:
            self.send_message(message)
        else:
            print('-'*40)
            print(message.as_string())
            print('-'*40)

    def get_subject(self):
        if self.subject:
            return self.render_text_template(self.subject)
        elif self.subject_template:
            return self.render_template(self.subject_template)
        else:
            return ''

    def get_body_text(self):
        if self.body_text:
            return self.render_text_template(self.body_text)
        elif self.body_text_template:
            return self.render_template(self.body_text_template)
        else:
            return ''

    def get_body_html(self):
        html = ''
        if self.body_html:
            html = transform(self.render_text_template(self.body_html))
        elif self.body_html_template:
            html = transform(self.render_template(self.body_html_template))
        return html

    def get_message(self):
        subject = self.get_subject()
        body_text = self.get_body_text()
        body_html = self.get_body_html()

        message = MIMEMultipart('alternative')
        message.set_charset('UTF-8')

        message['Subject'] = subject
        message['From'] = self.sender
        message['To'] = self._format_recipients(self.to)
        if self.cc:
            message['Cc'] = self._format_recipients(self.cc)
        if self.bcc:
            message['Bcc'] = self._format_recipients(self.bcc)
        if self.reply_to:
            message['reply-to'] = self.reply_to

        message.attach(MIMEText(body_text, 'plain'))
        if body_html:
            message.attach(MIMEText(body_html, 'html'))

        return message

    def send_message(self, message):
        raise NotImplementedError

    def _format_recipients(self, recipients):
        if isinstance(recipients, (list, tuple)):
            return ', '.join(recipients)
        else:
            return recipients



from __future__ import absolute_import
from .fixtures.descriptions import *

MONITOR_DESCRIPTIONS = [
    # --------------------------------------------------------------
    # monitor class                 description
    # --------------------------------------------------------------
    (NotDescriptedMonitor,          ''),
    (DocstringDescriptedMonitor,    'docstring monitor description'),
    (DescoratedDescriptedMonitor,   'decorator monitor description'),
    (DescoratedDescriptedMonitor2,  'decorator monitor description'),
]

SUITE_DESCRIPTIONS = [
    # --------------------------------------------------------------
    # suite class                   description
    # --------------------------------------------------------------
    (NotDescriptedSuite,            ''),
    (DocstringDescriptedSuite,      'docstring suite description'),
    (DescoratedDescriptedSuite,     'decorator suite description'),
    (DescoratedDescriptedSuite2,    'decorator suite description'),
]

METHOD_DESCRIPTIONS = [
    # ---------------------------------------------------------------------------------------------
    # monitor class                 method name                     description
    # ---------------------------------------------------------------------------------------------
    (DescriptedMethodsMonitor,      'test_not_descripted',          ''),
    (DescriptedMethodsMonitor,      'test_docstring_descripted',    'docstring method description'),
    (DescriptedMethodsMonitor,      'test_decorator_descripted',    'decorator method description'),
    (DescriptedMethodsMonitor,      'test_decorator_descripted2',   'decorator method description'),
]


def test_monitor_descriptions():
    for monitor_class, description in MONITOR_DESCRIPTIONS:
        assert monitor_class().monitor_description == description


def test_suite_descriptions():
    for suite_class, description in SUITE_DESCRIPTIONS:
        assert suite_class().description == description


def test_method_descriptions():
    for monitor_class, method_name, description in METHOD_DESCRIPTIONS:
        assert monitor_class(method_name).method_description == description



from __future__ import absolute_import
from spidermon.core.options import ActionOptions
from spidermon.decorators import OptionsDecorator


name = OptionsDecorator.set_value(ActionOptions, 'name')
description = OptionsDecorator.set_value(ActionOptions, 'description')




from __future__ import absolute_import
from spidermon import settings

from .fixtures.levels import *

HIGH = settings.MONITOR.LEVEL.HIGH
NORMAL = settings.MONITOR.LEVEL.NORMAL
LOW = settings.MONITOR.LEVEL.LOW


LEVEL_TESTS = [
    # ---------------------------------------------------------------------------------------
    # suite                     monitor/methods                                expected level
    # ---------------------------------------------------------------------------------------
    # Suite No Level
    (Suites.NoLevelSuite,       Monitors.NoLevelMonitor.NoLevelMethod,         NORMAL),
    (Suites.NoLevelSuite,       Monitors.NoLevelMonitor.HighLevelMethod,       HIGH),
    (Suites.NoLevelSuite,       Monitors.NoLevelMonitor.NormalLevelMethod,     NORMAL),
    (Suites.NoLevelSuite,       Monitors.NoLevelMonitor.LowLevelMethod,        LOW),

    (Suites.NoLevelSuite,       Monitors.HighLevelMonitor.NoLevelMethod,       HIGH),
    (Suites.NoLevelSuite,       Monitors.HighLevelMonitor.HighLevelMethod,     HIGH),
    (Suites.NoLevelSuite,       Monitors.HighLevelMonitor.NormalLevelMethod,   NORMAL),
    (Suites.NoLevelSuite,       Monitors.HighLevelMonitor.LowLevelMethod,      LOW),

    (Suites.NoLevelSuite,       Monitors.NormalLevelMonitor.NoLevelMethod,     NORMAL),
    (Suites.NoLevelSuite,       Monitors.NormalLevelMonitor.HighLevelMethod,   HIGH),
    (Suites.NoLevelSuite,       Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NoLevelSuite,       Monitors.NormalLevelMonitor.LowLevelMethod,    LOW),

    (Suites.NoLevelSuite,       Monitors.LowLevelMonitor.NoLevelMethod,        LOW),
    (Suites.NoLevelSuite,       Monitors.LowLevelMonitor.HighLevelMethod,      HIGH),
    (Suites.NoLevelSuite,       Monitors.LowLevelMonitor.NormalLevelMethod,    NORMAL),
    (Suites.NoLevelSuite,       Monitors.LowLevelMonitor.LowLevelMethod,       LOW),

    # Suite High
    (Suites.HighLevelSuite,     Monitors.NoLevelMonitor.NoLevelMethod,         HIGH),
    (Suites.HighLevelSuite,     Monitors.NoLevelMonitor.HighLevelMethod,       HIGH),
    (Suites.HighLevelSuite,     Monitors.NoLevelMonitor.NormalLevelMethod,     NORMAL),
    (Suites.HighLevelSuite,     Monitors.NoLevelMonitor.LowLevelMethod,        LOW),

    (Suites.HighLevelSuite,     Monitors.HighLevelMonitor.NoLevelMethod,       HIGH),
    (Suites.HighLevelSuite,     Monitors.HighLevelMonitor.HighLevelMethod,     HIGH),
    (Suites.HighLevelSuite,     Monitors.HighLevelMonitor.NormalLevelMethod,   NORMAL),
    (Suites.HighLevelSuite,     Monitors.HighLevelMonitor.LowLevelMethod,      LOW),

    (Suites.HighLevelSuite,     Monitors.NormalLevelMonitor.NoLevelMethod,     NORMAL),
    (Suites.HighLevelSuite,     Monitors.NormalLevelMonitor.HighLevelMethod,   HIGH),
    (Suites.HighLevelSuite,     Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.HighLevelSuite,     Monitors.NormalLevelMonitor.LowLevelMethod,    LOW),

    (Suites.HighLevelSuite,     Monitors.LowLevelMonitor.NoLevelMethod,        LOW),
    (Suites.HighLevelSuite,     Monitors.LowLevelMonitor.HighLevelMethod,      HIGH),
    (Suites.HighLevelSuite,     Monitors.LowLevelMonitor.NormalLevelMethod,    NORMAL),
    (Suites.HighLevelSuite,     Monitors.LowLevelMonitor.LowLevelMethod,       LOW),

    # Suite Normal
    (Suites.NormalLevelSuite,   Monitors.NoLevelMonitor.NoLevelMethod,         NORMAL),
    (Suites.NormalLevelSuite,   Monitors.NoLevelMonitor.HighLevelMethod,       HIGH),
    (Suites.NormalLevelSuite,   Monitors.NoLevelMonitor.NormalLevelMethod,     NORMAL),
    (Suites.NormalLevelSuite,   Monitors.NoLevelMonitor.LowLevelMethod,        LOW),

    (Suites.NormalLevelSuite,   Monitors.HighLevelMonitor.NoLevelMethod,       HIGH),
    (Suites.NormalLevelSuite,   Monitors.HighLevelMonitor.HighLevelMethod,     HIGH),
    (Suites.NormalLevelSuite,   Monitors.HighLevelMonitor.NormalLevelMethod,   NORMAL),
    (Suites.NormalLevelSuite,   Monitors.HighLevelMonitor.LowLevelMethod,      LOW),

    (Suites.NormalLevelSuite,   Monitors.NormalLevelMonitor.NoLevelMethod,     NORMAL),
    (Suites.NormalLevelSuite,   Monitors.NormalLevelMonitor.HighLevelMethod,   HIGH),
    (Suites.NormalLevelSuite,   Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.NormalLevelSuite,   Monitors.NormalLevelMonitor.LowLevelMethod,    LOW),

    (Suites.NormalLevelSuite,   Monitors.LowLevelMonitor.NoLevelMethod,        LOW),
    (Suites.NormalLevelSuite,   Monitors.LowLevelMonitor.HighLevelMethod,      HIGH),
    (Suites.NormalLevelSuite,   Monitors.LowLevelMonitor.NormalLevelMethod,    NORMAL),
    (Suites.NormalLevelSuite,   Monitors.LowLevelMonitor.LowLevelMethod,       LOW),

    # Suite Low
    (Suites.LowLevelSuite,      Monitors.NoLevelMonitor.NoLevelMethod,         LOW),
    (Suites.LowLevelSuite,      Monitors.NoLevelMonitor.HighLevelMethod,       HIGH),
    (Suites.LowLevelSuite,      Monitors.NoLevelMonitor.NormalLevelMethod,     NORMAL),
    (Suites.LowLevelSuite,      Monitors.NoLevelMonitor.LowLevelMethod,        LOW),

    (Suites.LowLevelSuite,      Monitors.HighLevelMonitor.NoLevelMethod,       HIGH),
    (Suites.LowLevelSuite,      Monitors.HighLevelMonitor.HighLevelMethod,     HIGH),
    (Suites.LowLevelSuite,      Monitors.HighLevelMonitor.NormalLevelMethod,   NORMAL),
    (Suites.LowLevelSuite,      Monitors.HighLevelMonitor.LowLevelMethod,      LOW),

    (Suites.LowLevelSuite,      Monitors.NormalLevelMonitor.NoLevelMethod,     NORMAL),
    (Suites.LowLevelSuite,      Monitors.NormalLevelMonitor.HighLevelMethod,   HIGH),
    (Suites.LowLevelSuite,      Monitors.NormalLevelMonitor.NormalLevelMethod, NORMAL),
    (Suites.LowLevelSuite,      Monitors.NormalLevelMonitor.LowLevelMethod,    LOW),

    (Suites.LowLevelSuite,      Monitors.LowLevelMonitor.NoLevelMethod,        LOW),
    (Suites.LowLevelSuite,      Monitors.LowLevelMonitor.HighLevelMethod,      HIGH),
    (Suites.LowLevelSuite,      Monitors.LowLevelMonitor.NormalLevelMethod,    NORMAL),
    (Suites.LowLevelSuite,      Monitors.LowLevelMonitor.LowLevelMethod,       LOW),
]


def test_levels():
    for suite, monitor, expected_level in LEVEL_TESTS:
        suite = suite()
        suite.add_monitor(monitor)
        assert suite.all_monitors[0].level == expected_level





from __future__ import absolute_import
import pytest

from spidermon.data import Data
from spidermon.exceptions import InvalidDataOperation

@pytest.fixture
def data():
    return Data(item_scraped_count=150)


def test_attribute_access(data):
    assert(data['item_scraped_count'] == 150)


def test_dictionary_access(data):
    assert(data.item_scraped_count == 150)


def test_attribute_set(data):
    with pytest.raises(InvalidDataOperation):
        data.item_scraped_count = 'some value'


def test_dictionary_set(data):
    with pytest.raises(InvalidDataOperation):
        data['item_scraped_count'] = 'some value'


def test_delete(data):
    with pytest.raises(InvalidDataOperation):
        del data['item_scraped_count']


def test_pop(data):
    with pytest.raises(InvalidDataOperation):
        data.pop('item_scraped_count', None)


def test_clear(data):
    with pytest.raises(InvalidDataOperation):
        data.clear()


def test_update(data):
    with pytest.raises(InvalidDataOperation):
        data.update({'item_scraped_count': 0})


def test_setdefault(data):
    with pytest.raises(InvalidDataOperation):
        data.setdefault('another_value', 0)



from __future__ import absolute_import
from spidermon.core.options import MonitorOptions
from spidermon import settings
from spidermon.decorators import OptionsDecorator, DecoratorWithAttributes


class LevelDecorator(DecoratorWithAttributes):
    name = 'level'
    attributes = {
        'high': OptionsDecorator.set_fixed_value(MonitorOptions, name, settings.MONITOR.LEVEL.HIGH),
        'normal': OptionsDecorator.set_fixed_value(MonitorOptions, name, settings.MONITOR.LEVEL.NORMAL),
        'low': OptionsDecorator.set_fixed_value(MonitorOptions, name, settings.MONITOR.LEVEL.LOW),
    }

name = OptionsDecorator.set_value(MonitorOptions, 'name')
description = OptionsDecorator.set_value(MonitorOptions, 'description')
order = OptionsDecorator.set_value(MonitorOptions, 'order')
level = LevelDecorator()



from __future__ import absolute_import
from collections import OrderedDict
import time
from six import itervalues

from spidermon import settings

from .items import ItemResult, MonitorResult, ActionResult


class Step(object):
    item_result_class = ItemResult
    successful_statuses = []
    error_statuses = []

    def __init__(self, name):
        self.name = name
        self._results = OrderedDict()
        self.start_time = 0
        self.finish_time = 0

    @property
    def time_taken(self):
        return self.finish_time - self.start_time

    @property
    def number_of_items(self):
        return len(self._results)

    def add_item(self, item):
        result = self.item_result_class(item)
        self._results[item] = result
        return result

    def start(self):
        self.start_time = time.time()

    def finish(self):
        self.finish_time = time.time()

    def items_for_status(self, status):
        return [result for item, result in self._results.items()
                if result.status == status]

    def items_for_statuses(self, statuses):
        items = []
        for status in statuses:
            items += self.items_for_status(status)
        return items

    @property
    def all_items(self):
        return list(itervalues(self._results))

    def get_infos(self):
        raise NotImplementedError

    def __getitem__(self, key):
        return self._results[key]

    @property
    def successful_results(self):
        results = []
        for successful_status in self.successful_statuses:
            results += self.items_for_status(successful_status)
        return results

    @property
    def error_results(self):
        results = []
        for error_status in self.error_statuses:
            results += self.items_for_status(error_status)
        return results

    @property
    def successful(self):
        return not self.has_errors

    @property
    def has_errors(self):
        return len(self.error_results) > 0


class MonitorStep(Step):
    item_result_class = MonitorResult
    successful_statuses = settings.MONITOR.STATUSES.SUCCESSFUL
    error_statuses = settings.MONITOR.STATUSES.ERROR

    def get_infos(self):
        return {
            'failures': len(self.items_for_status(settings.MONITOR.STATUS.FAILURE)),
            'errors': len(self.items_for_status(settings.MONITOR.STATUS.ERROR)),
            'skipped': len(self.items_for_status(settings.MONITOR.STATUS.SKIPPED)),
            'expected failures': len(self.items_for_status(settings.MONITOR.STATUS.EXPECTED_FAILURE)),
            'unexpected successes': len(self.items_for_status(settings.MONITOR.STATUS.UNEXPECTED_SUCCESS)),
        }


class ActionsStep(Step):
    item_result_class = ActionResult
    successful_statuses = settings.ACTION.STATUSES.SUCCESSFUL
    error_statuses = settings.ACTION.STATUSES.ERROR

    def get_infos(self):
        return {
            'errors': len(self.items_for_status(settings.ACTION.STATUS.ERROR)),
            'skipped': len(self.items_for_status(settings.ACTION.STATUS.SKIPPED)),
        }



# ----------------------------------
# Status
# ----------------------------------
UNDEFINED_STATUS = '?'

# ----------------------------------
# Monitors
# ----------------------------------
_MONITOR_STATUS_SUCCESS = 'OK'
_MONITOR_STATUS_FAILURE = 'FAIL'
_MONITOR_STATUS_SKIPPED = 'SKIPPED'
_MONITOR_STATUS_ERROR = 'ERROR'
_MONITOR_STATUS_EXPECTED_FAILURE = 'EXPECTED FAILURE'
_MONITOR_STATUS_UNEXPECTED_SUCCESS = 'UNEXPECTED SUCCESS'

_MONITOR_LEVEL_HIGH = 'HIGH'
_MONITOR_LEVEL_NORMAL = 'NORMAL'
_MONITOR_LEVEL_LOW = 'LOW'


class MONITOR:
    class STATUS:
        SUCCESS = _MONITOR_STATUS_SUCCESS
        FAILURE = _MONITOR_STATUS_FAILURE
        SKIPPED = _MONITOR_STATUS_SKIPPED
        ERROR = _MONITOR_STATUS_ERROR
        EXPECTED_FAILURE = _MONITOR_STATUS_EXPECTED_FAILURE
        UNEXPECTED_SUCCESS = _MONITOR_STATUS_UNEXPECTED_SUCCESS

    class STATUSES(object):
        ALL = (
            _MONITOR_STATUS_SUCCESS,
            _MONITOR_STATUS_FAILURE,
            _MONITOR_STATUS_SKIPPED,
            _MONITOR_STATUS_ERROR,
            _MONITOR_STATUS_EXPECTED_FAILURE,
            _MONITOR_STATUS_UNEXPECTED_SUCCESS,
        )
        SUCCESSFUL = (
            _MONITOR_STATUS_SUCCESS,
            _MONITOR_STATUS_SKIPPED,
            _MONITOR_STATUS_EXPECTED_FAILURE,
            _MONITOR_STATUS_UNEXPECTED_SUCCESS,
        )
        ERROR = (
            _MONITOR_STATUS_FAILURE,
            _MONITOR_STATUS_ERROR,
        )
        DEFAULT = UNDEFINED_STATUS

    class LEVEL:
        HIGH = _MONITOR_LEVEL_HIGH
        NORMAL = _MONITOR_LEVEL_NORMAL
        LOW = _MONITOR_LEVEL_LOW

    class LEVELS:
        ALL = (
            _MONITOR_LEVEL_HIGH,
            _MONITOR_LEVEL_NORMAL,
            _MONITOR_LEVEL_LOW,
        )
        DEFAULT = _MONITOR_LEVEL_NORMAL

    DEFAULT_NAME = None
    DEFAULT_DESCRIPTION = ''
    DEFAULT_ORDER = 1

# ----------------------------------
# Actions
# ----------------------------------
_ACTION_STATUS_SUCCESS = 'OK'
_ACTION_STATUS_ERROR = 'ERROR'
_ACTION_STATUS_SKIPPED = 'SKIPPED'


class ACTION:
    class STATUS:
        SUCCESS = _ACTION_STATUS_SUCCESS
        ERROR = _ACTION_STATUS_ERROR
        SKIPPED = _ACTION_STATUS_SKIPPED

    class STATUSES(object):
        ALL = (
            _ACTION_STATUS_SUCCESS,
            _ACTION_STATUS_ERROR,
            _ACTION_STATUS_SKIPPED,
        )
        SUCCESSFUL = (
            _ACTION_STATUS_SUCCESS,
            _ACTION_STATUS_SKIPPED,
        )
        ERROR = (
            _ACTION_STATUS_ERROR,
        )
        DEFAULT = UNDEFINED_STATUS

    DEFAULT_DESCRIPTION = ''


# ----------------------------------
# Steps
# ----------------------------------
_STEP_MONITORS = 'MONITORS'
_STEP_MONITORS_FINISHED = 'FINISHED ACTIONS'
_STEP_MONITORS_PASSED = 'PASSED ACTIONS'
_STEP_MONITORS_FAILED = 'FAILED ACTIONS'


class STEPS:
    MONITORS = _STEP_MONITORS
    MONITORS_FINISHED = _STEP_MONITORS_FINISHED
    MONITORS_PASSED = _STEP_MONITORS_PASSED
    MONITORS_FAILED = _STEP_MONITORS_FAILED
    ALL = (
        _STEP_MONITORS,
        _STEP_MONITORS_FINISHED,
        _STEP_MONITORS_PASSED,
        _STEP_MONITORS_FAILED,
    )
    MONITOR_RELATED = (
        _STEP_MONITORS,
    )
    ACTION_RELATED = (
        _STEP_MONITORS_FINISHED,
        _STEP_MONITORS_PASSED,
        _STEP_MONITORS_FAILED,
    )


from __future__ import absolute_import
import json
from schematics.models import Model
from schematics.types import URLType, StringType, BaseType


class TestValidator(Model):
    url = URLType(required=True)
    title = StringType()


class TreeValidator(Model):
    child = BaseType(required=True)


tree_schema = {
    "$schema": "http://json-schema.org/draft-04/schema",
    "required": [
        "child"
    ],
    "type": "object",
    "properties": {
        "child": {
            "type": "object"
        }
    }
}

test_schema = {
    "$schema": "http://json-schema.org/draft-04/schema",
    "required": [
        "url"
    ],
    "type": "object",
    "properties": {
        "url": {
            "type": "string"
        },
        "title": {
            "type": "string"
        },
    }
}

test_schema_string = json.dumps(test_schema)



from __future__ import absolute_import
import boto

from spidermon.exceptions import NotConfigured

from . import SendEmail


class SendSESEmail(SendEmail):
    aws_access_key = None
    aws_secret_key = None

    def __init__(self, aws_access_key=None, aws_secret_key=None, *args, **kwargs):
        super(SendSESEmail, self).__init__(*args, **kwargs)
        self.aws_access_key = aws_access_key or self.aws_access_key
        self.aws_secret_key = aws_secret_key or self.aws_secret_key
        if not self.fake and not self.aws_access_key:
            raise NotConfigured("You must provide the AWS Access Key.")
        if not self.fake and not self.aws_secret_key:
            raise NotConfigured("You must provide the AWS Secret Key.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super(SendSESEmail, cls).from_crawler_kwargs(crawler)
        kwargs.update({
            'aws_access_key': crawler.settings.get('SPIDERMON_AWS_ACCESS_KEY'),
            'aws_secret_key': crawler.settings.get('SPIDERMON_AWS_SECRET_KEY'),
        })
        return kwargs

    def send_message(self, message):
        session = boto.connect_ses(self.aws_access_key, self.aws_secret_key)
        session.send_raw_email(raw_message=message.as_string())



from __future__ import absolute_import
from spidermon.contrib.validation.translator import MessageTranslator
from spidermon.contrib.validation import messages


class JSONSchemaMessageTranslator(MessageTranslator):
    messages = {
        r"^.+ is a required property$":                         messages.MISSING_REQUIRED_FIELD,

        r"^.+ is not of type u?'array'$":                       messages.INVALID_ARRAY,
        r"^.+ is not of type u?'boolean'$":                     messages.INVALID_BOOLEAN,
        r"^.+ is not of type u?'integer'$":                     messages.INVALID_INT,
        r"^.+ is not of type u?'number'$":                      messages.INVALID_NUMBER,
        r"^.+ is not of type u?'object'$":                      messages.INVALID_OBJECT,
        r"^.+ is not of type u?'string'$":                      messages.INVALID_STRING,
        r"^.+ is not of type u?'null'$":                        messages.NOT_NULL,

        r"^.+ is not valid under any of the given schemas$":    messages.NOT_VALID_UNDER_ANY_SCHEMA,
        r"^.+ is valid under each of .+$":                      messages.VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS,

        r"^Additional items are not allowed .*$":               messages.TOO_MANY_ITEMS,
        r"^Additional properties are not allowed .*$":          messages.UNEXPECTED_FIELD,

        r"^.+ is a dependency of .+$":                          messages.MISSING_DEPENDENT_FIELD,
        r"^.* is not one of .+$":                               messages.VALUE_NOT_IN_CHOICES,

        r"^.* is not a 'date-time'$":                           messages.INVALID_DATETIME,
        r"^.* is not a 'email'$":                               messages.INVALID_EMAIL,
        r"^.* is not a 'ipv4'$":                                messages.INVALID_IPV4,
        r"^.* is not a 'ipv6'$":                                messages.INVALID_IPV6,
        r"^.* is not a 'hostname'$":                            messages.INVALID_HOSTNAME,
        r"^.* is not a 'url'$":                                 messages.INVALID_URL,
        r"^.* is not a 'uri'$":                                 messages.INVALID_URI,
        r"^.* is not a 'regex'$":                               messages.INVALID_REGEX,
        r"^.* is not a 'color'$":                               messages.INVALID_COLOR,

        r"^.* has too many properties$":                        messages.TOO_MANY_PROPERTIES,
        r"^.* does not have enough properties":                 messages.NOT_ENOUGH_PROPERTIES,

        r"^.* is greater than the maximum of .*$":              messages.NUMBER_TOO_HIGH,
        r"^.* is greater than or equal to the maximum of .*$":  messages.NUMBER_TOO_HIGH,

        r"^.* is less than the minimum of .*$":                 messages.NUMBER_TOO_LOW,
        r"^.* is less than or equal to the minimum of .*$":     messages.NUMBER_TOO_LOW,

        r"^.* is not a multiple of .*$":                        messages.NOT_MULTIPLE_OF,

        r"^.* is not allowed for .*$":                          messages.NOT_ALLOWED_VALUE,


        r"^.+ is too short$":                                   messages.FIELD_TOO_SHORT,
        r"^.+ is too long$":                                    messages.FIELD_TOO_LONG,

        r"^.+ does not match .*$":                              messages.REGEX_NOT_MATCHED,
        r"^.+ has non-unique elements$":                        messages.NOT_UNIQUE,
    }




from __future__ import absolute_import
from . import SendSlackMessage


class SendSlackMessageSpiderStarted(SendSlackMessage):
    message_template = 'slack/spider/notifier/start/message.jinja'
    include_attachments = False


class SendSlackMessageSpiderFinished(SendSlackMessage):
    message_template = 'slack/spider/notifier/finish/message.jinja'
    attachments_template = 'slack/spider/notifier/finish/attachments.jinja'
    include_ok_attachments = False
    include_error_attachments = True
    include_report_link = True
    report_index = 0

    def __init__(self,
                 include_ok_attachments=None,
                 include_error_attachments=None,
                 include_report_link=None,
                 report_index=None,
                 *args, **kwargs):
        super(SendSlackMessageSpiderFinished, self).__init__(*args, **kwargs)
        self.include_ok_attachments = include_ok_attachments or self.include_ok_attachments
        self.include_error_attachments = include_error_attachments or self.include_error_attachments
        self.include_report_link = include_report_link or self.include_report_link
        self.report_index = report_index or self.report_index

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super(SendSlackMessageSpiderFinished, cls).from_crawler_kwargs(crawler)
        kwargs.update({
            'include_ok_attachments': crawler.settings.get('SPIDERMON_SLACK_NOTIFIER_INCLUDE_OK_ATTACHMENTS'),
            'include_error_attachments': crawler.settings.get('SPIDERMON_SLACK_NOTIFIER_INCLUDE_ERROR_ATTACHMENTS'),
            'include_report_link': crawler.settings.get('SPIDERMON_SLACK_NOTIFIER_INCLUDE_REPORT_LINK'),
            'report_index': crawler.settings.get('SPIDERMON_SLACK_NOTIFIER_REPORT_INDEX'),
        })
        return kwargs

    def get_attachments(self):
        if (self.monitors_failed and self.include_error_attachments) or\
           (self.monitors_passed and self.include_ok_attachments):
            return super(SendSlackMessageSpiderFinished, self).get_attachments()
        else:
            return None

    def get_template_context(self):
        context = super(SendSlackMessageSpiderFinished, self).get_template_context()
        context.update({
            'include_ok_attachments': self.include_ok_attachments,
            'include_error_attachments': self.include_error_attachments,
            'include_report_link': self.include_report_link,
            'report_index': self.report_index,
        })
        return context


class SendSlackMessageSpiderRunning(SendSlackMessageSpiderFinished):
    message_template = 'slack/spider/notifier/periodic/message.jinja'
    attachments_template = 'slack/spider/notifier/periodic/attachments.jinja'



from __future__ import absolute_import
from spidermon import monitors, Monitor, MonitorSuite


# ----------------------------------
# Monitors
# ----------------------------------
class UnnamedMonitor(Monitor):
    def test_without_name(self):
        pass

    @monitors.name('A Test')
    def test_with_name(self):
        pass


@monitors.name('Class Monitor')
class NamedMonitor(Monitor):
    def test_without_name(self):
        pass

    @monitors.name('A Test')
    def test_with_name(self):
        pass


# ----------------------------------
# Child Suites
# ----------------------------------
class BaseChildSuite(MonitorSuite):
    monitors = [
        UnnamedMonitor,
        NamedMonitor,
        ('Instance Monitor', UnnamedMonitor),
        ('Instance Monitor', NamedMonitor),
    ]


class ChildUnnamedSuite(BaseChildSuite):
    pass


@monitors.name('The Child Suite')
class ChildNamedSuite(BaseChildSuite):
    pass


# ----------------------------------
# Top Suites
# ----------------------------------
class BaseTopSuite(MonitorSuite):
    monitors = [
        ChildUnnamedSuite,
        ChildNamedSuite,
        ('Instance Suite Name', ChildUnnamedSuite),
        ('Instance Suite Name', ChildNamedSuite),
    ]


class UnnamedTopSuite(BaseTopSuite):
    pass


@monitors.name('The Top Suite')
class NamedTopSuite(BaseTopSuite):
    pass



from __future__ import absolute_import
from spidermon.contrib.validation.translator import MessageTranslator
from spidermon.contrib.validation import messages


class SchematicsMessageTranslator(MessageTranslator):
    messages = {
        r"^Rogue field$":                                               messages.UNEXPECTED_FIELD,

        # BaseType
        r"^This field is required.$":                                   messages.MISSING_REQUIRED_FIELD,
        r"^Value must be one of .*\.$":                                 messages.VALUE_NOT_IN_CHOICES,

        # StringType
        r"^Couldn't interpret '.*' as string\.$":                       messages.INVALID_STRING,
        r"^String value is too long\.$":                                messages.FIELD_TOO_LONG,
        r"^String value is too short\.$":                               messages.FIELD_TOO_SHORT,
        r"^String value did not match validation regex\.$":             messages.REGEX_NOT_MATCHED,

        # DateTimeType
        r"^Could not parse .+\. Should be ISO ?8601(?: or timestamp)?\.$":  messages.INVALID_DATETIME,
        r"^Could not parse .+\. Valid formats: .+$":                    messages.INVALID_DATETIME,

        # DateType
        r"^Could not parse .+\. Should be ISO ?8601 \(YYYY-MM-DD\)\.$": messages.INVALID_DATE,

        # NumberType
        r"^.+ value should be greater than .+$":                        messages.NUMBER_TOO_LOW,
        r"^.+ value should be less than .+$":                           messages.NUMBER_TOO_HIGH,

        # IntType
        r"^Value '.*' is not int\.?$":                                  messages.INVALID_INT,

        # FloatType
        r"^Value '.*' is not float\.?$":                                messages.INVALID_FLOAT,

        # LongType
        r"^Value '.*' is not long\.?$":                                 messages.INVALID_LONG,

        # Decimalype
        r"^Number '.*' failed to convert to a decimal\.?$":             messages.INVALID_DECIMAL,
        r"^Value '.*' is not decimal\.?$":                              messages.INVALID_DECIMAL,
        r"^Value should be greater than .+$":                           messages.NUMBER_TOO_LOW,
        r"^Value should be less than .+$":                              messages.NUMBER_TOO_HIGH,

        # BooleanType
        r'^Must be either true or false\.$':                            messages.INVALID_BOOLEAN,

        # EmailType
        r"^Not a well[ -]formed email address\.$":                      messages.INVALID_EMAIL,

        # URLType
        r"^Not a well[ -]formed URL\.$":                                messages.INVALID_URL,

        # UUIDType
        r"^Couldn't interpret '.*' value as UUID\.$":                   messages.INVALID_UUID,

        # IPv4Type
        r"^Invalid IPv4 address$":                                      messages.INVALID_IPV4,

        # HashType
        r"^Hash value is wrong length\.$":                              messages.INVALID_HASH_LENGTH,
        r"^Hash value is not hexadecimal\.$":                           messages.INVALID_HASH,

        # ListType
        r"^Invalid list$":                                              messages.INVALID_LIST,
        r"^Could not interpret the value as a list$":                   messages.INVALID_LIST,
        r"^Please provide at least \d+ items?\.$":                      messages.LIST_TOO_SHORT,
        r"^Please provide no more than \d+ items?\.$":                  messages.LIST_TOO_LONG,

        # DictType
        r"^Only (?:dictionaries|mappings) may be used in a DictType$":  messages.INVALID_DICT,

        # DictType
        r"^Please use a mapping for this field or .+ "
        r"instance instead of .*\.$":                                   messages.INVALID_CHILD_CONTENT,
    }



from __future__ import absolute_import
from unittest import TestCase

from spidermon.data import Data
from spidermon import settings
from .options import MonitorOptions, MonitorOptionsMetaclass
import six


class Monitor(six.with_metaclass(MonitorOptionsMetaclass, TestCase)):
    def __init__(self, methodName='runTest', name=None):
        super(Monitor, self).__init__(methodName)
        self._name = name
        self._data = None
        self._parent = None
        self._init_method()

    @property
    def name(self):
        return '/'.join([self.monitor_name, self.method_name])

    @property
    def full_name(self):
        parts = []
        if self.parent and self.parent.full_name:
            parts.append(self.parent.full_name)
        parts.append(self.name)
        return '/'.join(parts)

    @property
    def level(self):
        return self.method_level or \
               self.monitor_level or \
               self.parent_level or \
               settings.MONITOR.LEVELS.DEFAULT

    @property
    def order(self):
        return self.method.options.order

    @property
    def monitor_name(self):
        return self._name or \
               self.options.name or \
               self.__class__.__name__

    @property
    def monitor_full_name(self):
        parts = []
        if self.parent and self.parent.full_name:
            parts.append(self.parent.full_name)
        parts.append(self.monitor_name)
        return '/'.join(parts)

    @property
    def monitor_description(self):
        return self.options.description or \
               self.__class__.__doc__ or \
               settings.MONITOR.DEFAULT_DESCRIPTION

    @property
    def monitor_level(self):
        return self.options.level

    @property
    def method(self):
        return getattr(self, self._testMethodName)

    @property
    def method_name(self):
        return self.method.options.name or \
               self._testMethodName

    @property
    def method_description(self):
        return self.method.options.description or \
               self.method.__func__.__doc__ or \
               settings.MONITOR.DEFAULT_DESCRIPTION

    @property
    def method_level(self):
        return self.method.options.level

    @property
    def parent(self):
        return self._parent

    @property
    def parent_level(self):
        return self.parent.level

    def set_parent(self, parent):
        self._parent = parent

    def init_data(self, data):
        self.data = data

    def debug_tree(self, level=0):
        return level*'\t' + repr(self) + '\n'

    def _init_method(self):
        MonitorOptions.add_or_create(self.method.__func__)

    def __repr__(self):
        return '<MONITOR:(%s) at %s>' % (self.name, hex(id(self)))

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(id(self))



from __future__ import absolute_import
import json
from jsonschema import validate

from spidermon import Monitor
from spidermon.core.options import MonitorOptions
from spidermon.python import Interpreter, schemas
from spidermon.exceptions import InvalidMonitor, NotConfigured
from spidermon import settings


class PythonExpressionsMonitor(Monitor):
    _classes_counter = 0
    _test_methods_counter = 0
    _test_methods_prefix = 'test'

    @classmethod
    def generate_class_name(cls):
        cls._classes_counter += 1
        return '%s%d' % (cls.__name__, cls._classes_counter)

    @classmethod
    def generate_method_name(cls):
        cls._test_methods_counter += 1
        return '%s_python_expression_%d' % (cls._test_methods_prefix,
                                            cls._test_methods_counter)

    def get_context_data(self):
        raise NotConfigured('Context data needs to be set up')


def create_monitor_class_from_json(monitor_json, monitor_class=None):
    monitor_dict = json.loads(monitor_json)
    validate(monitor_dict, schemas.MONITOR_SCHEMA)
    return create_monitor_class_from_dict(monitor_dict, monitor_class)


def create_monitor_class_from_dict(monitor_dict, monitor_class=None):
    tests = []
    for test in monitor_dict.get('tests', []):
        tests.append((
            test['expression'],
            test.get('name', None),
            test.get('description', None),
            test.get('fail_message', None),
        ))
    klass = _create_monitor_class(tests, monitor_class)
    klass.options.name = monitor_dict.get('name', settings.MONITOR.DEFAULT_NAME)
    klass.options.description = monitor_dict.get('description',
                                                 settings.MONITOR.DEFAULT_DESCRIPTION)
    return klass


def _create_monitor_class(expressions, monitor_class=None):
    monitor_class = monitor_class or PythonExpressionsMonitor
    if not issubclass(monitor_class, PythonExpressionsMonitor):
        msg = 'Python expressions monitors must subclass PythonExpressionsMonitor'
        raise InvalidMonitor(msg)
    klass = type(monitor_class.generate_class_name(), (monitor_class,), {})
    for e in expressions:
        if isinstance(e, tuple):
            method = _create_test_method(*e)
        else:
            method = _create_test_method(e)
        setattr(klass, monitor_class.generate_method_name(), method)
    return klass


def _create_test_method(expression, name=None, description=None, fail_reason=None):
    def _test_method(self):
        interpreter = Interpreter()
        context = self.get_context_data()
        result = interpreter.eval(expression, context=context)
        if result is not None:
            self.assertTrue(bool(result),
                            msg=('Expression not safisfied: "%s"' % expression)
                            if not fail_reason else
                            interpreter.eval(fail_reason, context=context))
    test_method = _test_method
    MonitorOptions.add_or_create(test_method)
    test_method.options.name = name or settings.MONITOR.DEFAULT_NAME
    test_method.options.description = description or settings.MONITOR.DEFAULT_DESCRIPTION
    return test_method




from __future__ import absolute_import
import re

from jsonschema.validators import Draft4Validator

from spidermon.contrib.validation.validator import Validator

from .translator import JSONSchemaMessageTranslator
from .formats import format_checker


REQUIRED_RE = re.compile("'(.+)' is a required property")


class JSONSchemaValidator(Validator):
    default_translator = JSONSchemaMessageTranslator()
    name = 'JSONSchema'

    def __init__(self, schema, translator=None, use_default_translator=True):
        super(JSONSchemaValidator, self).__init__(
            translator=translator,
            use_default_translator=use_default_translator,
        )
        self._schema = schema

    def _validate(self, data, strict=False):
        validator = Draft4Validator(
            schema=self._schema,
            format_checker=format_checker,
        )
        errors = validator.iter_errors(data)
        for e in errors:
            #print e
            absolute_path = list(e.absolute_path)
            required_match = REQUIRED_RE.search(e.message)
            if required_match:
                absolute_path.append(required_match.group(1))
            field_name = '.'.join([str(p) for p in absolute_path])
            self._add_errors({field_name: [e.message]})



from __future__ import absolute_import
import pytest

from spidermon import MonitorLoader
from spidermon.exceptions import InvalidMonitor

from .fixtures.cases import *
from .test_suites import check_suite


@pytest.fixture
def loader():
    return MonitorLoader()


def test_loading(loader):
    check_suite(
        suite=loader.load_suite_from_monitor(EmptyMonitor),
        expected_number_of_monitors=0
    )
    check_suite(
        suite=loader.load_suite_from_monitor(Monitor01),
        expected_number_of_monitors=3
    )
    check_suite(
        suite=loader.load_suite_from_monitor(Monitor02),
        expected_number_of_monitors=2
    )


def test_loading_errors(loader):
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(None)
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(10)
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(object)


def test_testcase_names(loader):
    _check_testcase_names(
        loader=loader,
        monitor_class=EmptyMonitor,
        expected_names=[]
    )
    _check_testcase_names(
        loader=loader,
        monitor_class=Monitor01,
        expected_names=[
            'test_a',
            'test_b',
            'test_c',
        ]
    )
    _check_testcase_names(
        loader=loader,
        monitor_class=Monitor02,
        expected_names=[
            'test_d',
            'test_e',
        ]
    )


def _check_testcase_names(loader, monitor_class, expected_names):
    names = loader.get_testcase_names(monitor_class)
    assert names == expected_names




from __future__ import absolute_import
import hashlib

from boto.s3.connection import S3Connection
from boto.s3.key import Key

from spidermon.exceptions import NotConfigured

from . import CreateReport


DEFAULT_S3_REGION_ENDPOINT = 's3.amazonaws.com'
DEFAULT_S3_CONTENT_TYPE = 'text/html'
URL_SECRET_KEY = 'The secret to life the universe and everything'


class S3Uploader(object):
    def __init__(self, aws_key, aws_secret):
        self.connection = S3Connection(
            aws_access_key_id=aws_key,
            aws_secret_access_key=aws_secret)

    def upload_from_file(self, source_filename, s3_bucket, s3_filename, headers=None, make_public=False):
        with open(source_filename, 'r') as f:
            self._upload_with_method(
                bucket=s3_bucket,
                method_name='set_contents_from_file',
                filename=s3_filename,
                content=f,
                headers=headers,
                make_public=make_public,
            )

    def upload_from_content(self, content, s3_bucket, s3_filename, headers=None, make_public=False):
        self._upload_with_method(
            bucket=s3_bucket,
            method_name='set_contents_from_string',
            filename=s3_filename,
            content=content,
            headers=headers,
            make_public=make_public,
        )

    def _upload_with_method(self, bucket, method_name, filename, content, headers=None, make_public=False):
        # Get bucket without validation (Needed to be used with credentials w/o listing perms)
        bucket = self.connection.get_bucket(bucket, validate=False)
        f = Key(bucket)
        f.key = filename
        getattr(f, method_name)(content, headers=headers)
        if make_public:
            f.make_public()


class CreateS3Report(CreateReport):
    aws_access_key = None
    aws_secret_key = None
    s3_bucket = None
    s3_filename = None
    s3_region_endpoint = DEFAULT_S3_REGION_ENDPOINT
    make_public = True
    content_type = DEFAULT_S3_CONTENT_TYPE

    def __init__(self, aws_access_key=None, aws_secret_key=None,
                 s3_bucket=None, s3_filename=None, s3_region_endpoint=None,
                 make_public=False, content_type=None,
                 *args, **kwargs):
        super(CreateS3Report, self).__init__(*args, **kwargs)

        self.aws_access_key = aws_access_key or self.aws_access_key
        self.aws_secret_key = aws_secret_key or self.aws_secret_key
        self.s3_bucket = s3_bucket or self.s3_bucket
        self.s3_region_endpoint = s3_region_endpoint or self.s3_region_endpoint
        self.s3_filename = s3_filename or self.s3_filename
        self.make_public = make_public or self.make_public
        self.content_type = content_type or self.content_type
        if not self.aws_access_key:
            raise NotConfigured("You must provide the AWS Access Key.")
        if not self.aws_secret_key:
            raise NotConfigured("You must provide the AWS Secret Key.")
        if not self.s3_bucket:
            raise NotConfigured("You must define the s3 bucket.")
        if not self.s3_filename:
            raise NotConfigured("You must define the s3 filename.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super(CreateS3Report, cls).from_crawler_kwargs(crawler)
        kwargs.update({
            'aws_access_key': crawler.settings.get('SPIDERMON_AWS_ACCESS_KEY'),
            'aws_secret_key': crawler.settings.get('SPIDERMON_AWS_SECRET_KEY'),
            's3_bucket': crawler.settings.get('SPIDERMON_REPORT_S3_BUCKET'),
            's3_filename': crawler.settings.get('SPIDERMON_REPORT_S3_FILENAME'),
            's3_region_endpoint': crawler.settings.get('SPIDERMON_REPORT_S3_REGION_ENDPOINT'),
            'make_public': crawler.settings.get('SPIDERMON_REPORT_S3_MAKE_PUBLIC'),
            'content_type': crawler.settings.get('SPIDERMON_REPORT_S3_CONTENT_TYPE'),
        })
        return kwargs

    def after_render_report(self):
        s3 = S3Uploader(self.aws_access_key, self.aws_secret_key)
        s3.upload_from_content(
            content=self.report,
            s3_bucket=self.s3_bucket,
            s3_filename=self.get_s3_filename(),
            headers={'Content-Type': self.content_type},
            make_public=self.make_public,
        )

    def get_s3_filename(self):
        return 'reports/{secret}/{filename}'.format(
            secret=self.get_url_secret(),
            filename=self.render_text_template(self.s3_filename),
        )

    def get_s3_report_url(self):
        return 'https://{region}/{bucket}/{filename}'.format(
            region=self.s3_region_endpoint,
            bucket=self.s3_bucket,
            filename=self.get_s3_filename(),
        )

    def get_url_secret(self):
        secret = URL_SECRET_KEY
        if self.data.job:
            secret += str(self.data.job.key.split('/')[0])
        return hashlib.md5(secret.encode()).hexdigest()

    def get_meta(self):
        report_url = self.get_s3_report_url()
        return {
            'reports_links': self.data.meta.get('reports', []) + [report_url]
        }




from __future__ import absolute_import
import six
import ast

from spidermon.exceptions import InvalidExpression


class Interpreter(object):

    ast_allowed_nodes = (
        'expr', 'name', 'load', 'call', 'store',
        'str', 'unicode', 'num', 'list', 'dict', 'set', 'tuple',  # Data types
        'nameconstant',  # language constant (None)
        'unaryop', 'usub',  # Unary arithmetic operators
        'binop', 'add', 'sub', 'div', 'mult', 'mod', 'pow', 'floordiv',  # Binary arithmetic operators
        'compare', 'eq', 'noteq', 'gt', 'lt', 'gte', 'lte',  # Comparison operators
        'bitand', 'bitor', 'bitxor', 'invert', 'lshift', 'rshift',  # Bitwise operators
        'boolop', 'and', 'or', 'not',  # Logical operators
        'in', 'notin',  # Membership operators
        'is', 'isnot',  # Identity operators
        'ifexp',  # Inline if statement
        'subscript', 'index', 'slice', 'extslice',  # Subscripting
        'listcomp', 'setcomp', 'dictcomp',  'generatorexp', 'comprehension',  # Comprehensions
        'attribute',  # Attribute access
    )

    allowed_objects = (
        str, six.text_type,  # strings
        int, float, complex,  # numbers
        list, dict, set, tuple,  # sequences
        type(None), bool  # others
    )
    try:
        # Py2 compatibility
        allowed_objects += (long,)
    except NameError:
        pass

    def check(self, expression):
        if not isinstance(expression, six.string_types):
            raise InvalidExpression('Python expressions must be defined as strings')
        if not expression:
            raise InvalidExpression('Empty python expression')

        try:
            tree = ast.parse(expression)
        except SyntaxError as e:
            raise e

        if not tree.body:
            raise InvalidExpression('Empty python expression')
        elif len(tree.body) > 1:
            raise InvalidExpression('Python expressions must be a single line expression')

        start_node = tree.body[0]
        if not isinstance(start_node, ast.Expr):
            raise InvalidExpression("Python string must be an expression: '%s' found" %
                                    start_node.__class__.__name__)

        self._check_node(start_node)

    def eval(self, expression, context=None, check=True):
        if check:
            self.check(expression)
        return eval(expression, context)

    def _check_node(self, node):
        if isinstance(node, list):
            self._check_node_list(node)
        elif isinstance(node, ast.AST):
            if not self._is_allowed_ast_node(node):
                self._raise_not_allowed_node(node)
            self._check_node_fields(node)
        elif not isinstance(node, self.allowed_objects):
            self._raise_not_allowed_node(node)

    def _check_node_list(self, node_list):
        for node in node_list:
            self._check_node(node)

    def _check_node_fields(self, node):
        for field in [f for _, f in ast.iter_fields(node)]:
            self._check_node(field)

    def _is_allowed_ast_node(self, node):
        return node.__class__.__name__.lower() in self.ast_allowed_nodes

    def _raise_not_allowed_node(self, node):
        raise InvalidExpression("'%s' definition not allowed in python expressions" %
                                node.__class__.__name__)



from __future__ import absolute_import
from slugify import slugify


STATS_DEFAULT_VALIDATION_PREFIX = 'spidermon/validation'


class NAMES:
    ITEMS = 'items'
    DROPPED = 'dropped'
    FIELDS = 'fields'
    ERRORS = 'errors'
    VALIDATORS = 'validators'


class ValidationStatsManager(object):
    def __init__(self, stats, prefix=None, slugify=True):
        self.stats = stats
        self.prefix = prefix or STATS_DEFAULT_VALIDATION_PREFIX
        self.slugify = slugify

    def add_validator(self, type, class_name):
        self.stats.inc_value(self._get_stats_name(NAMES.VALIDATORS))
        self.stats.set_value(self._get_stats_name(NAMES.VALIDATORS, type, class_name), True)

    def add_field_error(self, field, error):
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS))
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS, error))
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS, error) + '/' + field)

    def add_fields(self, count):
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS), count=count)

    def add_item(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS))

    def add_dropped_item(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS, NAMES.DROPPED))

    def add_item_with_errors(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS, NAMES.ERRORS))

    def _get_stats_name(self, *names):
        return '/'.join([self.prefix] + list([self._get_name(n) for n in names]))

    def _get_name(self, name):
        return slugify(text=name, separator='_').lower() if self.slugify else name



from __future__ import absolute_import

from scrapy import signals
from scrapy.utils.misc import load_object
from scrapy.exceptions import NotConfigured
from twisted.internet.task import LoopingCall

from spidermon import MonitorSuite
from spidermon.contrib.scrapy.runners import SpiderMonitorRunner
from spidermon.utils.hubstorage import hs
from spidermon.python import factory
from spidermon.python.monitors import ExpressionsMonitor


class Spidermon(object):

    def __init__(self, crawler,
                 spider_opened_suites=None, spider_closed_suites=None,
                 spider_opened_expression_suites=None, spider_closed_expression_suites=None,
                 expressions_monitor_class=None,
                 periodic_suites=None):
        if not crawler.settings.getbool('SPIDERMON_ENABLED'):
            raise NotConfigured
        self.crawler = crawler

        self.spider_opened_suites = [self.load_suite(s) for s in spider_opened_suites or []]
        self.spider_opened_suites += [self.load_expression_suite(s, expressions_monitor_class)
                                      for s in spider_opened_expression_suites or []]

        self.spider_closed_suites = [self.load_suite(s) for s in spider_closed_suites or []]
        self.spider_closed_suites += [self.load_expression_suite(s, expressions_monitor_class)
                                      for s in spider_closed_expression_suites or []]

        self.periodic_suites = periodic_suites or {}
        self.periodic_tasks = {}

    def load_suite(self, suite_to_load):
        try:
            suite_class = load_object(suite_to_load)
        except Exception as e:
            raise e  # TO-DO
        if not issubclass(suite_class, MonitorSuite):
            raise Exception  # TO-DO
        return suite_class(crawler=self.crawler)

    def load_expression_suite(self, suite_to_load, monitor_class=None):
        if monitor_class:
            monitor_class = load_object(monitor_class)
        else:
            monitor_class = ExpressionsMonitor
        monitor = factory.create_monitor_class_from_dict(
            monitor_dict=suite_to_load,
            monitor_class=monitor_class
        )
        suite = MonitorSuite(crawler=self.crawler)
        suite.add_monitor(monitor)
        return suite

    @classmethod
    def from_crawler(cls, crawler):
        ext = cls(
            crawler=crawler,
            spider_opened_suites=crawler.settings.getlist('SPIDERMON_SPIDER_OPEN_MONITORS'),
            spider_closed_suites=crawler.settings.getlist('SPIDERMON_SPIDER_CLOSE_MONITORS'),
            spider_opened_expression_suites=crawler.settings.getlist('SPIDERMON_SPIDER_OPEN_EXPRESSION_MONITORS'),
            spider_closed_expression_suites=crawler.settings.getlist('SPIDERMON_SPIDER_CLOSE_EXPRESSION_MONITORS'),
            expressions_monitor_class=crawler.settings.get('SPIDERMON_EXPRESSIONS_MONITOR_CLASS'),
            periodic_suites=crawler.settings.getdict('SPIDERMON_PERIODIC_MONITORS'),
        )
        crawler.signals.connect(ext.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)
        return ext

    def spider_opened(self, spider):
        self._run_suites(spider, self.spider_opened_suites)
        self.periodic_tasks[spider] = []
        for suite, time in self.periodic_suites.items():
            task = LoopingCall(self._run_periodic_suites, spider, [suite])
            self.periodic_tasks[spider].append(task)
            task.start(time, now=False)

    def spider_closed(self, spider):
        self._run_suites(spider, self.spider_closed_suites)
        for task in self.periodic_tasks[spider]:
            task.stop()

    def _run_periodic_suites(self, spider, suites):
        suites = [self.load_suite(s) for s in suites]
        self._run_suites(spider, suites)

    def _run_suites(self, spider, suites):
        data = self._generate_data_for_spider(spider)
        for suite in suites:
            runner = SpiderMonitorRunner(spider=spider)
            runner.run(suite, **data)

    def _generate_data_for_spider(self, spider):
        return {
            'stats': self.crawler.stats.get_stats(spider),
            'crawler': self.crawler,
            'spider': spider,
            'job': hs.job if hs.available else None,
        }



from __future__ import absolute_import
from .fixtures.ordering import *

SUITE_SEQUENCES = [
    # ------------------------------------------------------------------------------------------------------------
    # monitor sequence                                      expected sequence
    # ------------------------------------------------------------------------------------------------------------

    # Unordered: should follow same insertion order (FIFO)
    ([Unordered.A, Unordered.B],                            [Unordered.A, Unordered.B]),
    ([Unordered.B, Unordered.A],                            [Unordered.B, Unordered.A]),
    ([Unordered.C, Unordered.D],                            [Unordered.C, Unordered.D]),
    ([Unordered.D, Unordered.C],                            [Unordered.D, Unordered.C]),
    ([Unordered.A, Unordered.B, Unordered.C, Unordered.D],  [Unordered.A, Unordered.B, Unordered.C, Unordered.D]),
    ([Unordered.D, Unordered.C, Unordered.B, Unordered.A],  [Unordered.D, Unordered.C, Unordered.B, Unordered.A]),
    ([Unordered.C, Unordered.B, Unordered.A, Unordered.D],  [Unordered.C, Unordered.B, Unordered.A, Unordered.D]),
    ([Unordered.A, Unordered.A, Unordered.B, Unordered.B],  [Unordered.A, Unordered.A, Unordered.B, Unordered.B]),
    ([Unordered.C, Unordered.C, Unordered.B, Unordered.B],  [Unordered.C, Unordered.C, Unordered.B, Unordered.B]),

    # Ordered: should follow decorators order (A, B, C, D)
    ([Ordered.A, Ordered.B],                                [Ordered.A, Ordered.B]),
    ([Ordered.B, Ordered.A],                                [Ordered.A, Ordered.B]),
    ([Ordered.C, Ordered.D],                                [Ordered.C, Ordered.D]),
    ([Ordered.D, Ordered.C],                                [Ordered.C, Ordered.D]),
    ([Ordered.A, Ordered.B, Ordered.C, Ordered.D],          [Ordered.A, Ordered.B, Ordered.C, Ordered.D]),
    ([Ordered.D, Ordered.C, Ordered.B, Ordered.A],          [Ordered.A, Ordered.B, Ordered.C, Ordered.D]),
    ([Ordered.C, Ordered.B, Ordered.A, Ordered.D],          [Ordered.A, Ordered.B, Ordered.C, Ordered.D]),
    ([Ordered.A, Ordered.A, Ordered.B, Ordered.B],          [Ordered.A, Ordered.A, Ordered.B, Ordered.B]),
    ([Ordered.C, Ordered.C, Ordered.B, Ordered.B],          [Ordered.B, Ordered.B, Ordered.C, Ordered.C]),
]

METHOD_SEQUENCES = [
    # --------------------------------------------------------------
    # monitor class                 expected methods sequence
    # --------------------------------------------------------------
    (UnorderedMethodsMonitor,       ['test_a', 'test_b', 'test_c']),
    (OrderedMethodsMonitor,         ['test_c', 'test_b', 'test_a']),
    (EqualOrderedMethodsMonitor,    ['test_a', 'test_b', 'test_c']),
]


def test_suite_ordering():
    for monitors_sequence, expected_sequence in SUITE_SEQUENCES:
        suite = MonitorSuite()
        suite.add_monitors(monitors_sequence)
        sequence = [_extract_monitor_class(m) for m in suite]
        assert sequence == expected_sequence


def test_method_ordering():
    for monitor_class, expected_sequence in METHOD_SEQUENCES:
        suite = MonitorSuite()
        suite.add_monitor(monitor_class)
        sequence = [m.method_name for m in suite.all_monitors]
        assert sequence == expected_sequence


def _extract_monitor_class(suite):
    if suite.__class__ == MonitorSuite:
        return _extract_monitor_class(suite._tests[0])
    else:
        return suite.__class__



from __future__ import absolute_import
from spidermon.contrib.stats.counters import PercentCounter, DictPercentCounter, AttributeDictPercentCounter
from spidermon.contrib.stats.analyzer import StatsAnalyzer
from spidermon.contrib.scrapy.stats import STATS_DEFAULT_VALIDATION_PREFIX

from .stats import StatsMonitorMixin


class MetaDictPercentCounter(DictPercentCounter):
    def add_values(self, key, subkey, value):
        if key not in self._dict:
            self._create_item(key)
        self[key].add_value(subkey, value)


class ErrorsDictPercentCounter(AttributeDictPercentCounter):
    __attribute_dict_name__ = 'fields'


class ErrorsInfo(MetaDictPercentCounter):
    __items_class__ = ErrorsDictPercentCounter


class FieldErrorsDictPercentCounter(AttributeDictPercentCounter):
    __attribute_dict_name__ = 'errors'


class FieldErrorsInfo(MetaDictPercentCounter):
    __items_class__ = FieldErrorsDictPercentCounter

    def __init__(self, fields_count, items_count):
        super(FieldErrorsInfo, self).__init__(items_count)
        self._fields_count = fields_count

    @property
    def count(self):
        return self._fields_count


class ItemsInfo(object):
    def __init__(self, items_count, items_with_errors, items_dropped):
        self.count = items_count
        self.errors = PercentCounter(count=items_with_errors, total=items_count)
        self.dropped = PercentCounter(count=items_dropped, total=items_count)


class ValidationInfo(object):
    def __init__(self, stats, prefix=None):
        self.analyzer = StatsAnalyzer(stats=stats, prefix=prefix or STATS_DEFAULT_VALIDATION_PREFIX)

        # items
        items_count = sum(self.analyzer.search('items$').values())
        items_with_errors_count = sum(self.analyzer.search('items/errors$').values())
        items_dropped_count = sum(self.analyzer.search('items/dropped$').values())
        self.items = ItemsInfo(
            items_count=items_count,
            items_with_errors=items_with_errors_count,
            items_dropped=items_dropped_count)

        # errors & fields
        fields_count = sum(self.analyzer.search('fields$').values())
        self.errors = ErrorsInfo(items_count)
        self.fields = FieldErrorsInfo(fields_count=fields_count, items_count=items_count)

        field_errors = self.analyzer.search('fields/errors/([^/]+)$', include_matches=True)
        for _, error in field_errors.values():
            field_errors_per_field = self.analyzer.search('fields/errors/%s/([^/]+)$' % error, include_matches=True)
            for count, field in field_errors_per_field.values():
                self.errors.add_values(key=error, subkey=field, value=count)
                self.fields.add_values(key=field, subkey=error, value=count)


class ValidationMonitorMixin(StatsMonitorMixin):
    @property
    def validation(self):
        if not hasattr(self, '_validation'):
            self._validation = ValidationInfo(self.stats)
        return self._validation

    def check_missing_required_fields(self, field_names=None, allowed_count=0):
        if not field_names:
            missing_count = self.validation.errors['missing_required_field'].count
            self._check_missing_required_count(missing_count, allowed_count)
        else:
            for field_name in field_names:
                self.check_missing_required_field(field_name, allowed_count)

    def check_missing_required_field(self, field_name, allowed_count=0):
        missing_count = self.validation.fields[field_name].errors['missing_required_field'].count
        self._check_missing_required_count(missing_count, allowed_count)

    def _check_missing_required_count(self, missing_count, allowed_count):
        self.assertLessEqual(
            missing_count,
            allowed_count,
            msg='{count} required field{plural} are missing!{threshold_info}'.format(
                count=missing_count,
                plural='' if missing_count == 1 else 's',
                threshold_info=(' (maximum allowed %d)' % allowed_count) if allowed_count > 0 else '',
            )
        )

    def check_missing_required_fields_percent(self, field_names=None, allowed_percent=0):
        if not field_names:
            missing_percent = self.validation.errors['missing_required_field'].percent
            self._check_missing_required_percent(missing_percent, allowed_percent)
        else:
            for field_name in field_names:
                self.check_missing_required_field_percent(field_name, allowed_percent)

    def check_missing_required_field_percent(self, field_name, allowed_percent=0):
        missing_percent = self.validation.fields[field_name].errors['missing_required_field'].percent
        self._check_missing_required_percent(missing_percent, allowed_percent)

    def _check_missing_required_percent(self, missing_percent, allowed_percent=0):
        self.assertLessEqual(
            missing_percent,
            allowed_percent,
            msg='{percent}% of required fields are missing!{threshold_info}'.format(
                percent=missing_percent*100,
                threshold_info=(' (maximum allowed %.0f%%)' % (allowed_percent*100)) if allowed_percent > 0 else '',
            )
        )

    def check_fields_errors(self, field_names=None, errors=None, allowed_count=0):
        if not field_names:
            errors_count = self.validation.errors.count
            self._check_field_errors(errors_count, allowed_count)
        else:
            for field_name in field_names:
                self.check_field_errors(field_name, errors, allowed_count)

    def check_field_errors(self, field_name, errors=None, allowed_count=0):
        if errors:
            errors_count = sum([self.validation.fields[field_name].errors[e].count for e in errors])
        else:
            errors_count = self.validation.fields[field_name].errors.count
        self._check_field_errors(errors_count, allowed_count)

    def _check_field_errors(self, errors_count, allowed_count):
        self.assertLessEqual(
            errors_count,
            allowed_count,
            msg='{count} field{count_plural} {verb} validation errors!{threshold_info}'.format(
                count=errors_count,
                count_plural='' if errors_count == 1 else 's',
                verb='has' if errors_count == 1 else 'have',
                threshold_info=(' (maximum allowed %d)' % allowed_count) if allowed_count > 0 else '',
            )
        )

    def check_fields_errors_percent(self, field_names=None, errors=None, allowed_percent=0):
        if not field_names:
            errors_percent = self.validation.errors.percent
            self._check_field_errors_percent(errors_percent, allowed_percent)
        else:
            for field_name in field_names:
                self.check_field_errors_percent(field_name, errors, allowed_percent)

    def check_field_errors_percent(self, field_name, errors=None, allowed_percent=0):
        if errors:
            errors_percent = sum([self.validation.fields[field_name].errors[e].percent for e in errors])
        else:
            errors_percent = self.validation.fields[field_name].errors.percent
        self._check_field_errors_percent(errors_percent, allowed_percent)


    def _check_field_errors_percent(self, errors_percent, allowed_percent):
        self.assertLessEqual(
            errors_percent,
            allowed_percent,
            msg='{percent}% of fields have validation errors!{threshold_info}'.format(
                percent=errors_percent*100,
                threshold_info=(' (maximum allowed %.0f%%)' % (allowed_percent*100)) if allowed_percent > 0 else '',
            )
        )



from setuptools import setup, find_packages

test_requirements = [
    'pytest>=2.7.0',
    'tox'
]

setup(
    name='spidermon',
    version='1.8.0',
    url='https://github.com/scrapinghub/spidermon',
    author='Scrapinghub',
    author_email='info@scrapinghub.com',
    description=('Spidermon is a framework to build monitors for Scrapy '
                 'spiders.'),
    long_description=('Spidermon is a framework to build monitors for Scrapy '
                      'spiders.'),
    license='BSD',
    packages=find_packages(),
    package_data={'spidermon': ['VERSION']},
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'six>=1.9.0',
    ],
    tests_require=test_requirements,
    extras_require={
        # Specific monitors and tools to support notifications and reports
        'monitoring': [
            'scrapy',
            'Jinja2',
            'slackclient',
            'boto',
            'premailer'
        ],
        # Data validation
        'validation': [
            'jsonschema',
            'schematics',
            'python-slugify',
            'strict-rfc3339'
        ],
        # Tools to run the tests
        'tests': test_requirements,
        # Tools to build and publish the documentation
        'docs': [
            'sphinx',
            'sphinx-rtd-theme',
            's3cmd'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Scrapy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Monitoring',
    ],
)
